/*
 * Copyright (C) 2001 the MCatS project
 *
 * This file is part of the MCatS, Modular Catalog System.
 *
 * MCatS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MCatS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <dlfcn.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>
#include "config.h"
#include "mcats.h"

int usage();
struct ModuleStruct * get_list_of_plugins( char *path, struct ModuleStruct *current_list_of_plugins, char * );
struct ModuleStruct * search_for_plugins( char * );
void list_plugins();

static struct MainStruct mainstructure =
{
	0,
	NULL,
	-1,
	NULL,
	-1,
	NULL,
	NULL,
	NULL
};


int
usage()
{
	printf("usage: mcats [options]\n"
			"options:\n"
			"   -i, --input plugin     Select a plugin for reading\n"
			"   -o, --output plugin    Select a plugin for writing\n"
			"   -l, --list             List plugins\n"
			"   -v, --version          Print the version of MCatS and exit.\n"
			"   -h, --help             Print this help and exits\n\n");
}


void
version()
{
	printf( "MCatS version %s\n", VERSION );
}

struct ModuleStruct *
search_for_plugins( char *modulename )
{
	struct ModuleStruct *first_list_of_plugins;
	struct ModuleStruct *current_list_of_plugins;
	char path[255] = "";
	char *homedir;
	char pwd[255] = "";
	int loop = 0;

	/* Here we allocate the first struct to store module information to. */
	/* We do it here so we can keep track of the first allocated structure */
	first_list_of_plugins = (struct ModuleStruct *) malloc( sizeof( struct ModuleStruct) );
	current_list_of_plugins = first_list_of_plugins;

	/* Now the home directory */
	if( ( homedir = getenv("HOME")) != NULL )
	{
		sprintf(path, "%s/%s", homedir, MCATS_PLUGIN_DIR);
		current_list_of_plugins = get_list_of_plugins(path, current_list_of_plugins, modulename );
	}

	/* Scan for system wide plugins first */
	sprintf(path, "%s", SYSTEM_PLUGIN_DIR);
	current_list_of_plugins = get_list_of_plugins(path, current_list_of_plugins, modulename );

	/* returns pointer to first allocated structure */
	return first_list_of_plugins;
}

struct ModuleStruct *
get_list_of_plugins( char *path, struct ModuleStruct *current_list_of_plugins, char *modulename )
{
	DIR *dir;
	struct dirent *dentry;
	struct ModuleStruct *new_list_of_plugins;
	void *handler;			// Handler for the loading of .so
	PluginStruct *(*load_plugin)();		// Pointer to load_plugin function
	PluginStruct *load_plugin_module;		// Store the struct from .so
	char module_file[255] = "";		// Full path/file to .so file
	char *so;
	int returnvalue = 0;

	/* Go through the module dirs and see which is MCatS compatible */
	if( (dir = opendir(path)) != NULL )
	{
		while( (dentry = readdir(dir)) != NULL )
		{
			if( dentry->d_name[0] != '.'
					&& dentry->d_ino > 0
					&& (so = strrchr(dentry->d_name, '.')) != NULL
					&& strcmp(so, ".so") == 0 )
			{
				sprintf(module_file, "%s/%s", path, dentry->d_name);

				handler = dlopen( module_file, DL_LAZY );
				if( handler )
				{
					load_plugin = dlsym( handler, "load_plugin" );
					load_plugin_module = (*load_plugin)();
					if( load_plugin_module && load_plugin_module->id != 0 )
					{
						/* Store the information from the module into the current struct */
						sprintf( current_list_of_plugins->filename, "%s", module_file );
						sprintf( current_list_of_plugins->shortname, "%s", load_plugin_module->shortname );
						sprintf( current_list_of_plugins->description, "%s", load_plugin_module->description );
						current_list_of_plugins->id = load_plugin_module->id;
						current_list_of_plugins->check_input_args = load_plugin_module->check_input_args;
						current_list_of_plugins->check_output_args = load_plugin_module->check_output_args;
						current_list_of_plugins->input = load_plugin_module->input;
						current_list_of_plugins->output = load_plugin_module->output;

						/* Allocate some space for next module */
						new_list_of_plugins = (struct ModuleStruct *)
							malloc( sizeof( struct ModuleStruct ) );
						/* Let the old struct know of the new struct */
						current_list_of_plugins->next = new_list_of_plugins;

						/*		if( modulename != NULL )
								if( strcmp( modulename, current_list_of_plugins->shortname ) == 0 )
								{
								returnvalue = dlclose( handler );
								closedir(dir);
								current_list_of_plugins = new_list_of_plugins;
								current_list_of_plugins->next = NULL;
								return current_list_of_plugins;
								}
						 */
						/* Make the new one the current one */
						current_list_of_plugins = new_list_of_plugins;
						current_list_of_plugins->next = NULL;
					}
					returnvalue = dlclose(handler);
					/* Should it come a warning if the module wasn't closed ok? */
				}
			}
		}
		closedir(dir);
	}

	return current_list_of_plugins;
}

void
list_plugins()
{
	struct ModuleStruct *first_list_of_plugins;
	struct ModuleStruct *current_list_of_plugins;
	struct ModuleStruct *next_list_of_plugins;

	/* Get's pointer to first structure */
	first_list_of_plugins = search_for_plugins(NULL);

	/* Display info about modules */
	current_list_of_plugins = first_list_of_plugins;

	printf( "\nList of available modules:\n" );
	printf("Type\tName\t\tDescription\n");

	while(current_list_of_plugins->next)
	{
		/* Print out types */
		if( current_list_of_plugins->id & MODULE_INPUT )
			printf( "i" );
		if( current_list_of_plugins->id & MODULE_OUTPUT )
			printf( "o" );

		if( strlen( current_list_of_plugins->shortname ) < 8 )
			printf("\t%s\t\t%s\n", current_list_of_plugins->shortname, current_list_of_plugins->description);
		else
			printf("\t%s\t%s\n", current_list_of_plugins->shortname, current_list_of_plugins->description);
		current_list_of_plugins = current_list_of_plugins->next;
	}

	printf( "\nDescription:\n" );
	printf( "i = Input Module\no = Output Module\n\n" );

	/* A rutine for deallocolating any memory used */
	current_list_of_plugins = first_list_of_plugins;
	while(current_list_of_plugins)
	{
		next_list_of_plugins = current_list_of_plugins->next;
		free(current_list_of_plugins);
		current_list_of_plugins = next_list_of_plugins;
	}
}

int
check_input_args( int argc, char *argv[], char input_plugin[15] )
{
	struct ModuleStruct *first_list_of_plugins = 0;
	struct ModuleStruct *current_list_of_plugins = 0;
	struct ModuleStruct *next_list_of_plugins = 0;
	PluginStruct *(*load_plugin)();		// Pointer to load_plugin function
	PluginStruct *load_plugin_module;		// Store the struct from .so
	void *handler;		// Handler for the loading of .so
	int ok_args = -3;

	/* Get's pointer to first structure */
	first_list_of_plugins = search_for_plugins(NULL);

	/* Display info about modules */
	current_list_of_plugins = first_list_of_plugins;
	while(current_list_of_plugins->next)
	{
		if( strcmp( current_list_of_plugins->shortname, input_plugin ) == 0 )
		{
#ifdef __DEBUG__
			fprintf( stderr, "DEBUG: Opening library !%s!\n", current_list_of_plugins->filename );
#endif
			handler = dlopen( current_list_of_plugins->filename, DL_LAZY );
			if( handler )
			{
				load_plugin = dlsym( handler, "load_plugin" );
				load_plugin_module = (*load_plugin)();
				current_list_of_plugins->check_input_args = load_plugin_module->check_input_args;

				if( load_plugin_module->id & MODULE_INPUT )
					ok_args = (*current_list_of_plugins->check_input_args)( argc, argv );
				dlclose(handler);
			}
		}
		current_list_of_plugins = current_list_of_plugins->next;
	}

	if( ok_args == -3 )
		fprintf( stderr, "WARNING: Didn't find the input module '%s'\n", input_plugin );

	/* A rutine for deallocolating any memory used */
	current_list_of_plugins = first_list_of_plugins;
	while(current_list_of_plugins)
	{
		next_list_of_plugins = current_list_of_plugins->next;
		free(current_list_of_plugins);
		current_list_of_plugins = next_list_of_plugins;
	}
	return ok_args;
}

int
check_output_args( int argc, char *argv[], char output_plugin[15] )
{
	struct ModuleStruct *first_list_of_plugins = 0;
	struct ModuleStruct *current_list_of_plugins = 0;
	struct ModuleStruct *next_list_of_plugins = 0;
	PluginStruct *(*load_plugin)();		// Pointer to load_plugin function
	PluginStruct *load_plugin_module;		// Store the struct from .so
	void *handler;			// Handler for the loading of .so
	int ok_args = -3;

	/* Get's pointer to first structure */
	first_list_of_plugins = search_for_plugins(NULL);

	/* Display info about modules */
	current_list_of_plugins = first_list_of_plugins;
	while(current_list_of_plugins->next)
	{
		if( strcmp( current_list_of_plugins->shortname, output_plugin ) == 0 )
		{
#ifdef __DEBUG__
			fprintf( stderr, "DEBUG: Opening library !%s!\n", current_list_of_plugins->filename );
#endif
			handler = dlopen( current_list_of_plugins->filename, DL_LAZY );
			if( handler )
			{
				load_plugin = dlsym( handler, "load_plugin" );
				load_plugin_module = (*load_plugin)();
				current_list_of_plugins->check_output_args = load_plugin_module->check_output_args;

				if( load_plugin_module->id & MODULE_OUTPUT )
					ok_args = (*current_list_of_plugins->check_output_args)( argc, argv );
#ifdef __DEBUG__
				fprintf( stderr, "DEBUG: check_output_args returned: %d\n", ok_args );
#endif
				dlclose(handler);
			}
		}
		current_list_of_plugins = current_list_of_plugins->next;
	}

	if( ok_args == -3 )
		fprintf( stderr, "WARNING: Didn't find the output module '%s'\n", output_plugin );

	/* A rutine for deallocolating any memory used */
	current_list_of_plugins = first_list_of_plugins;
	while(current_list_of_plugins)
	{
		next_list_of_plugins = current_list_of_plugins->next;
		free(current_list_of_plugins);
		current_list_of_plugins = next_list_of_plugins;
	}

	return ok_args;
}

struct VolumeStruct
do_input( char input_plugin[15] )
{
	struct ModuleStruct *first_list_of_plugins;
	struct ModuleStruct *current_list_of_plugins;
	struct ModuleStruct *next_list_of_plugins;
	struct VolumeStruct volumestructure;
	struct FileStruct *filestructure;
	void *handler;		// Handler for the loading of .so

	/* Get's pointer to first structure */
	first_list_of_plugins = search_for_plugins(input_plugin);

	/* Display info about modules */
	current_list_of_plugins = first_list_of_plugins;
	while(current_list_of_plugins->next)
	{
		if( strcmp( current_list_of_plugins->shortname, input_plugin ) == 0 )
		{
#ifdef __DEBUG__
			fprintf( stderr, "DEBUG: before dlopen, errno: %d\n", errno );
#endif
			handler = dlopen( current_list_of_plugins->filename, DL_LAZY );
#ifdef __DEBUG__
			fprintf( stderr, "DEBUG: after dlopen, errno: %d\n", errno );
#endif
			if( handler )
			{
				volumestructure = (*current_list_of_plugins->input)( &mainstructure );
#ifdef __DEBUG__
				fprintf( stderr, "DEBUG: before dlclose, errno: %d\n", errno );
#endif
				dlclose(handler);
#ifdef __DEBUG__
				fprintf( stderr, "DEBUG: after dlclose, errno: %d\n", errno );
#endif
			}
		}
		current_list_of_plugins = current_list_of_plugins->next;
	}

	if( volumestructure.filestructure == NULL )
		return volumestructure; //return NULL;

	return volumestructure;
}

int
do_output( char output_plugin[15], struct VolumeStruct volumestructure )
{
	struct ModuleStruct *first_list_of_plugins;
	struct ModuleStruct *current_list_of_plugins;
	struct ModuleStruct *next_list_of_plugins;
	void *handler;		// Handler for the loading of .so
	int ret_ok = 0;

	/* Get's pointer to first structure */
	first_list_of_plugins = search_for_plugins(output_plugin);

	/* Display info about modules */
	current_list_of_plugins = first_list_of_plugins;
	while(current_list_of_plugins->next)
	{
		if( strcmp( current_list_of_plugins->shortname, output_plugin ) == 0 )
		{
			handler = dlopen( current_list_of_plugins->filename, DL_LAZY );
			if( handler )
			{
				ret_ok = (*current_list_of_plugins->output)( &mainstructure, volumestructure );
				dlclose(handler);
				break;
			}
		}
		current_list_of_plugins = current_list_of_plugins->next;
	}

	if( volumestructure.filestructure == NULL )
		return 0;

	/* A rutine for deallocolating any memory used */

	current_list_of_plugins = first_list_of_plugins;
	while(current_list_of_plugins)
	{
		next_list_of_plugins = current_list_of_plugins->next;
		free(current_list_of_plugins);
		current_list_of_plugins = next_list_of_plugins;
	}

	return 0;
}


int
main( int argc, char *argv[] )
{
	struct VolumeStruct volumestructure;
	char input_plugin[15] = "";		// Name of the input plugin
	char output_plugin[15] = "";		// Name of the output plugin
	int listplugins = 0;			// Does the user want to list plugins
	int loop = 0;				// Standard loop variable
	int input_ok = -1;
	int output_ok = -1;

	if( argc < 2 )
	{
		usage();
		exit(0);
	}

	for( loop = 1; loop < argc; loop++ )

		if( ( strcmp( argv[loop], "-i" ) == 0
					|| strcmp( argv[loop], "--input" ) == 0 )
					&& argc > loop + 1 )
			strcpy( input_plugin, argv[++loop] );

		else if( ( strcmp( argv[loop], "-o" ) == 0
					|| strcmp( argv[loop], "--output" ) == 0 )
					&& argc > loop + 1 )
			strcpy( output_plugin, argv[++loop] );

		else if( strcmp( argv[loop], "-l" ) == 0
				|| strcmp( argv[loop], "--list" ) == 0 )
			listplugins = 1;

		else if( ( strcmp( argv[loop], "-v" ) ==0
					|| strcmp( argv[loop], "--version" ) == 0 )
					&& strlen( input_plugin ) < 1
					&& strlen( output_plugin ) < 1 )
		{
			version();
			exit(0);
		}

		else if( ( strcmp( argv[loop], "-h" ) == 0
					|| strcmp( argv[loop], "--help" ) == 0 )
					&& strlen( input_plugin ) < 1
					&& strlen( output_plugin ) < 1 )
		{
			usage();
			exit(0);
		}

	if( listplugins == 1 )
	{
		list_plugins();
		exit(0);
	}

	/* Check if the arguments was correct */
	if( strlen( input_plugin ) > 0 )
		input_ok = check_input_args( argc, argv, input_plugin );
	if( strlen( output_plugin ) > 0 )
		output_ok = check_output_args( argc, argv, output_plugin );

	if( input_ok == 0 && output_ok == 0 )
	{
		mainstructure.argc = argc;
		mainstructure.argv = argv;
		do {
			volumestructure = do_input( input_plugin );
			if( volumestructure.filestructure != NULL )
			{
				do_output( output_plugin, volumestructure );
			}
		} while( ( mainstructure.input_argc != -1 || mainstructure.output_argc != -1 ) && volumestructure.filestructure != NULL );
	}

	return 0;
}
