/*
 * Copyright (C) 2001 the MCatS project
 *
 * This file is part of the MCatS, Modular Catalog System.
 *
 * MCatS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MCatS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA	02111-1307, USA
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <fstab.h>
#include <sys/stat.h>
#include <getopt.h>
#ifdef LINUX
	#include <linux/limits.h>
	#include <sys/vfs.h>
	#include <mntent.h>
	#include <linux/cdrom.h>
	#include <sys/fcntl.h>
#else
	#include <sys/syslimits.h>
	#include <sys/param.h>
	#include <sys/mount.h>
	#include <isofs/cd9660/cd9660_mount.h>
#endif
#include <dirent.h>
#include <mcats.h>

#define SHORT_NAME "cdinput"
#define CDINPUT_ARGS_MOUNTPOINT 1
#define CDINPUT_ARGS_DEVICE 2
#define CDINPUT_ARGS_FILESYSTEM 4
#define CDINPUT_ARGS_CDLABEL 8
#define CDINPUT_ARGS_AUTOMOUNT 16
#define CDINPUT_ARGS_EJECT 32
#define CDINPUT_ARGS_VERSION 64
#define CDINPUT_ARGS_HELP 128

struct CDInputArgs {
	char cddevice[PATH_MAX + 1];
	char mountpoint[PATH_MAX + 1];
	char filesystem[256];
	char cdlabel[33];
	int Options;
};

void usage()
{
	printf("usage: mcats -i cdinput [options]\n"
		"options:\n"
		"   --mountpoint path    Mount point specified in fstab\n"
		"   --device dev         CD-Rom device\n"
		"   --filesystem fs      Filesystem to use\n"
		"   --cdlabel name       Name of CD-Rom volume\n"
		"   --automount          Don't try to mount or unmount cd CD-Rom\n"
		"   --eject              Eject CD-Rom when finished\n"
		"   --version            Prints the version of cdinput module and exits\n"
		"   --help               Prints this help and exits\n\n");
}

void version()
{
	printf( "cdinput version %s\n", CDINPUT_VERSION );
}

struct FileStruct *
read_files_on_cd( char *path, struct FileStruct *current_file_in_struct, char *init_path )
{
	struct FileStruct *new_file_in_struct;
	DIR *dir;
	struct dirent *dentry;
	struct stat fileinfo;
	char file[PATH_MAX + 1] = "";
	int result = 0;

	if( (dir = opendir(path)) != NULL )
	{
		while( (dentry = readdir(dir)) != NULL )
		{
			sprintf(file, "%s/%s", path, dentry->d_name);
			if( strcmp( dentry->d_name, "." ) != 0 &&
				strcmp( dentry->d_name, ".." ) != 0 )
			{
				if( lstat( file, &fileinfo ) == -1 )
				{
					fprintf(stderr, "WARNING: Bogues!!!!\n");
				}
				else if(S_ISLNK (fileinfo.st_mode) )
				{
					fprintf(stderr, "WARNING: A link on a cdrom???\n");
				}
				else if(S_ISDIR (fileinfo.st_mode) )
				{
					sprintf( current_file_in_struct->filename, "%s", dentry->d_name );
					if( strlen( path ) == strlen( init_path ) )
						sprintf( current_file_in_struct->path, "/" );
					else
						sprintf( current_file_in_struct->path, "%s/", (path + strlen(init_path)) );
					current_file_in_struct->size = fileinfo.st_size;
					current_file_in_struct->mode = fileinfo.st_mode;
					current_file_in_struct->owner = fileinfo.st_uid;
					current_file_in_struct->group = fileinfo.st_gid;
					new_file_in_struct = ( struct FileStruct *)
																					malloc( sizeof( struct FileStruct ) );
					new_file_in_struct->description = NULL;
					current_file_in_struct->next = new_file_in_struct;
					current_file_in_struct = new_file_in_struct;
					new_file_in_struct = read_files_on_cd( file, current_file_in_struct, init_path );
					current_file_in_struct = new_file_in_struct;
				} else {
					sprintf( current_file_in_struct->filename, "%s", dentry->d_name );
					if( strlen( path ) == strlen( init_path ) )
						sprintf( current_file_in_struct->path, "/");
					else
						sprintf( current_file_in_struct->path, "%s/", (path + strlen(init_path)) );
					current_file_in_struct->size = fileinfo.st_size;
					current_file_in_struct->mode = fileinfo.st_mode;
					current_file_in_struct->owner = fileinfo.st_uid;
					current_file_in_struct->group = fileinfo.st_gid;
					new_file_in_struct = ( struct FileStruct *)
																					malloc( sizeof( struct FileStruct ) );
					new_file_in_struct->description = NULL;
					current_file_in_struct->next = new_file_in_struct;
					current_file_in_struct = new_file_in_struct;
				}
			}
		}
	} else {
		fprintf(stderr, "ERROR: Couldn't open dir???\n");
	}
	closedir(dir);
	return current_file_in_struct;
}

struct VolumeStruct
read_cd( struct MainStruct *mainstructure )
{
	FILE *cddev;
	struct CDInputArgs cdinput_args;
	struct VolumeStruct volumestructure;
	struct FileStruct *first_file_in_struct = NULL;
	struct FileStruct *current_file_in_struct = NULL;
	struct FileStruct *next_file_in_struct = NULL;
	struct fstab *fs_info;
#ifndef LINUX
	struct iso_args cd_args;
#else
	FILE *fstabfile;
	struct mntent *mntentstruct;
	char mountstring[255] = "";
#endif
	struct statfs cdinfo;
	char cddevice[32] = "";
	char filesystem[32] = "";
	char mountpoint[255] = "";
	char first_file_on_cd[255] = "";	// Path to first file on cd (.)
	char cdbuf[1024] = "";
	int ok_args = 0;
	int status = 0;
	int loop = 0;


	/* Reset the volumestructure */
	volumestructure.name[0] = 0;
	volumestructure.publisher[0] = 0;
	volumestructure.preparer[0] = 0;
	volumestructure.application[0] = 0;
	volumestructure.f_type;
	volumestructure.f_oflags;
	volumestructure.f_bsize;
	volumestructure.f_iosize;
	volumestructure.f_blocks;
	volumestructure.f_bfree;
	volumestructure.f_bavail;
	volumestructure.f_files;
	volumestructure.f_ffree;
	volumestructure.f_fsid;
	volumestructure.f_owner;
	volumestructure.f_flags;
	volumestructure.f_syncwrites;
	volumestructure.f_asyncwrites;
	volumestructure.f_spare[1];
	volumestructure.f_fstypename[0];
	volumestructure.f_mntonname[0];
	volumestructure.f_mntfromname[0];
	/* *** */

	ok_args = read_args( mainstructure->argc, mainstructure->argv, &cdinput_args );

#ifdef LINUX

	if( cdinput_args.Options & CDINPUT_ARGS_MOUNTPOINT 
			&& !( cdinput_args.Options & CDINPUT_ARGS_DEVICE )
			&& !( cdinput_args.Options & CDINPUT_ARGS_FILESYSTEM ) )
	{
		/* Read the fstab file to see witch device the cdrom is */
		/* This is later used when reading the CDLabel */
		fstabfile = setmntent( "/etc/fstab", "r" );
		if( fstabfile != NULL )
		{
			mntentstruct = getmntent( fstabfile );
			while( mntentstruct != NULL )
			{
				if( strcmp( cdinput_args.mountpoint, mntentstruct->mnt_dir ) == 0 )
				{
					strcpy( cddevice, mntentstruct->mnt_fsname );
					strcpy( filesystem, mntentstruct->mnt_type );
					break;
				}
				mntentstruct = getmntent( fstabfile );
			}
			fclose( fstabfile );
		}
		if( strlen( cddevice ) == 0 || strlen( filesystem ) == 0 )
		{
			fprintf( stderr, "ERROR: Couldn't find mountpoint %s in /etc/fstab", cdinput_args.mountpoint );

			volumestructure.filestructure = first_file_in_struct;
			return volumestructure;
		}
		strcpy( mountpoint, cdinput_args.mountpoint );
		sprintf( mountstring, "mount %s", cdinput_args.mountpoint );
	} else if( cdinput_args.Options & CDINPUT_ARGS_DEVICE )
	{
		sprintf( cddevice, "%s", cdinput_args.cddevice );
		sprintf( mountpoint, "%s", cdinput_args.mountpoint );
		sprintf( filesystem, "%s", cdinput_args.filesystem );
		if( cdinput_args.Options & CDINPUT_ARGS_FILESYSTEM )
			sprintf( mountstring, "mount -t %s %s %s", cdinput_args.filesystem,
			cdinput_args.cddevice, cdinput_args.mountpoint );
		else
			sprintf( mountstring, "mount %s %s", cdinput_args.cddevice,
					cdinput_args.mountpoint );
	}

#else

	if( cdinput_args.Options & CDINPUT_ARGS_MOUNTPOINT 
			&& !( cdinput_args.Options & CDINPUT_ARGS_DEVICE )
			&& !( cdinput_args.Options & CDINPUT_ARGS_FILESYSTEM ) )
	{
		fs_info = getfsfile( cdinput_args.mountpoint );
		sprintf( cddevice, "%s", fs_info->fs_spec );
		sprintf( filesystem, "%s", fs_info->fs_vfstype );
		sprintf( mountpoint, "%s", cdinput_args.mountpoint );
	} else if( cdinput_args.Options & CDINPUT_ARGS_DEVICE )
	{
		sprintf( cddevice, "%s", cdinput_args.cddevice );
		sprintf( mountpoint, "%s", cdinput_args.mountpoint );
		if( cdinput_args.Options & CDINPUT_ARGS_FILESYSTEM )
			sprintf( filesystem, "%s", cdinput_args.filesystem );
		else
			sprintf( filesystem, "%s", MOUNT_CD9660 );
	}

#endif

	/* If cdlabel have been set on commandline, put it into structure */
	if( cdinput_args.Options & CDINPUT_ARGS_CDLABEL )
		strncpy( volumestructure.name, cdinput_args.cdlabel, 32 );

	/* Try to read the name and other info raw from the CD-ROM */
	if( strlen( cddevice ) > 0 && !( cdinput_args.Options & CDINPUT_ARGS_CDLABEL )
			&& !( cdinput_args.Options & CDINPUT_ARGS_AUTOMOUNT ) )
	{
		/* Check CDLabel if present */
		cddev = fopen( cddevice, "rb" );

		if( cddev == NULL )
		{
			fprintf(stderr, "ERROR: Couldn't open %s for reading\n", cddevice);
			volumestructure.filestructure = NULL;
			return volumestructure;
		}

		if( fseek( cddev, 32768, SEEK_SET ) < 0 )
		{
			fprintf(stderr, "ERROR: Error reading CD\n");
			fclose( cddev );
			volumestructure.filestructure = NULL;
			return volumestructure;
		}

		if( fread( cdbuf, sizeof( cdbuf ), 1, cddev ) == 0 )
		{
			fprintf(stderr, "ERROR: Error reading CD\n");
			fclose( cddev );
			volumestructure.filestructure = NULL;
			return volumestructure;
		}

		if( strncmp( cdbuf, "\001CD001\001", 7 ) != 0 )
		{
			fprintf(stderr, "WARNING: It's not a ISO9660 CD\n");
/*			fclose( cddev );
			volumestructure.filestructure = NULL;
			return volumestructure;*/
		}
		else
		{
			/* NULL terminate in case any entry is full length */
			volumestructure.name[32] = 0;
			volumestructure.publisher[128] = 0;
			volumestructure.preparer[128] = 0;
			volumestructure.application[128] = 0;

			/* Start by reading Volume ID */
			for( loop = 40; loop < 72; loop++ )
				if( cdbuf[loop] == 32 )
				{
					volumestructure.name[loop-40] = 0;
					continue;
				}
				else
					volumestructure.name[loop-40] = cdbuf[loop];

			/* Continue by reading Publisher */
			for( loop = 318; loop < 446; loop++ )
				if( cdbuf[loop] == 32 )
				{
					volumestructure.publisher[loop-318] = 0;
					continue;
				}
			 else
					volumestructure.publisher[loop-318] = cdbuf[loop];

			/* Continue by reading Preparer */
			for( loop = 446; loop < 574; loop++ )
				if( cdbuf[loop] == 32 )
				{
					volumestructure.preparer[loop-446] = 0;
					continue;
				}
				else
					volumestructure.preparer[loop-446] = cdbuf[loop];

			/* End by reading Application */
			for( loop = 574; loop < 702; loop++ )
				if( cdbuf[loop] == 32 )
				{
					volumestructure.application[loop-574] = 0;
					continue;
				}
				else
					volumestructure.application[loop-574] = cdbuf[loop];

		}

		/* Does a REAL cleanup so we can mount the cddevice correctly */
		fflush( cddev );
#ifndef LINUX
		fpurge( cddev );
#endif
		fclose( cddev );
	}

#ifdef LINUX

	/* Finaly mount the CD to read filestructure */
	if( ( cdinput_args.Options & CDINPUT_ARGS_MOUNTPOINT
				|| ( cdinput_args.Options & CDINPUT_ARGS_MOUNTPOINT
				&& cdinput_args.Options & CDINPUT_ARGS_DEVICE ) )
			&& !( cdinput_args.Options & CDINPUT_ARGS_AUTOMOUNT ) )
	{
		status = mount( cddevice, mountpoint, filesystem, MS_RDONLY, NULL );
	}

#else

	/* Finaly mount the CD to read filestructure */
	if( ( cdinput_args.Options & CDINPUT_ARGS_MOUNTPOINT
				|| ( cdinput_args.Options & CDINPUT_ARGS_MOUNTPOINT
				&& cdinput_args.Options & CDINPUT_ARGS_DEVICE ) )
			&& !( cdinput_args.Options & CDINPUT_ARGS_AUTOMOUNT ) )
	{
		cd_args.export.ex_flags = MNT_EXRDONLY;
		cd_args.fspec = cddevice;
		cd_args.export.ex_root = -2;
		status = mount( filesystem, mountpoint, MNT_RDONLY, &cd_args );
	}

#endif

	/* If it didn't mount correctly and it wasn't already mounted, return */
	if( status != 0 && errno != EBUSY )
	{
		fprintf( stderr, "ERROR: Problems mounting CD-Rom (%s on %s using %s): ", cddevice, mountpoint, filesystem);
		perror( strerror( errno ) );

		volumestructure.filestructure = first_file_in_struct;
		return volumestructure;
	}

	sprintf( first_file_on_cd, "%s", mountpoint );
	if( statfs( first_file_on_cd, &cdinfo ) != -1 )
	{
		/* Sucessfully got info about the filesystem */
		volumestructure.f_type = cdinfo.f_type;
#ifdef LINUX
/* FIXME: Is there a way to set those variables on Linux anyway? */
		volumestructure.f_oflags = 0;
		volumestructure.f_iosize = 0;
		volumestructure.f_flags = 0;
		volumestructure.f_syncwrites = 0;
		volumestructure.f_asyncwrites = 0;
		volumestructure.f_owner = 0;
		volumestructure.f_fstypename[0] = 0;
		volumestructure.f_mntonname[0] = 0;
		volumestructure.f_mntfromname[0] = 0;
#else
		volumestructure.f_oflags = cdinfo.f_oflags;
		volumestructure.f_iosize = cdinfo.f_iosize;
		volumestructure.f_flags = cdinfo.f_flags;
		volumestructure.f_syncwrites = cdinfo.f_syncwrites;
		volumestructure.f_asyncwrites = cdinfo.f_asyncwrites;
		volumestructure.f_owner = cdinfo.f_owner;
		sprintf( volumestructure.f_fstypename, "%s", cdinfo.f_fstypename );
		sprintf( volumestructure.f_mntonname, "%s", cdinfo.f_mntonname );
		sprintf( volumestructure.f_mntfromname, "%s", cdinfo.f_mntfromname );
#endif
		volumestructure.f_bsize = cdinfo.f_bsize;
		volumestructure.f_blocks = cdinfo.f_blocks;
		volumestructure.f_bfree = cdinfo.f_bfree;
		volumestructure.f_bavail = cdinfo.f_bavail;
		volumestructure.f_files = cdinfo.f_files;
		volumestructure.f_ffree = cdinfo.f_ffree;
		volumestructure.f_fsid = cdinfo.f_fsid;
		/* volumestructure.f_spare = volumestructure.f_spare; */ /* Not used */
	}
	first_file_in_struct = (struct FileStruct *) malloc( sizeof( struct FileStruct ) );
	first_file_in_struct->description = NULL;

	current_file_in_struct = read_files_on_cd( cdinput_args.mountpoint, first_file_in_struct, cdinput_args.mountpoint );

	if( current_file_in_struct == NULL )
	{
		fprintf( stderr, "WARNING: No files on cd\n" );
		free( first_file_in_struct );
		first_file_in_struct = NULL; //return NULL;
	}

	if( !( cdinput_args.Options & CDINPUT_ARGS_AUTOMOUNT ) )
/* Unmount mounted CD-Rom */
#ifdef LINUX
	status = umount( mountpoint );
#else
	status = unmount( mountpoint, NULL );
#endif

	if( status != 0 )
	{
		fprintf( stderr, "WARNING: Problems unmounting CD-Rom (mounted on %s): ", mountpoint);
		perror( strerror( errno ) );
	}

	if( cdinput_args.Options & CDINPUT_ARGS_AUTOMOUNT )
		/* Eject the CD-Rom */
#ifdef LINUX
	{
		int fd = open( cddevice, O_RDONLY|O_NONBLOCK );
		if( fd == -1 )
			fprintf( stderr, "WARNING: Problems ejecting CD-Rom" );
		else
		{
			status = ioctl( fd, CDROMEJECT );
			close( fd );
		}
	}
#else
		status = 0;
#endif

	if( status != 0 )
	{
		fprintf( stderr, "WARNING: Problems ejecting CD-Rom: " );
		perror( strerror( errno ) );
	}

	volumestructure.filestructure = first_file_in_struct;
	return volumestructure;
}

int
read_args( int argc, char *argv[], struct CDInputArgs *cdinput_args )
{
	int c;
	int stop_reading_args = 0;
	int loop = 0;
	/* Set getopt's index pointer */
	while( strcmp( argv[loop++], "-i" ) != 0 )
		;
	optind = loop - 1;

	/* Clear the structure so we can se if it was filled */
	strcpy( cdinput_args->cddevice, "" );
	strcpy( cdinput_args->mountpoint, "" );
	strcpy( cdinput_args->filesystem, "" );
	strcpy( cdinput_args->cdlabel, "" );
	cdinput_args->Options = 0;

	while(1)
	{
		int option_index = 0;

		static struct option cdinput_options[] =
		{
			{"input", 1, 0, 0},
			{"mountpoint", 1, 0, 1},
			{"device", 1, 0, 2},
			{"filesystem", 1, 0, 3},
			{"cdlabel", 1, 0, 4},
			{"automount", 0, 0, 5},
			{"eject", 0, 0, 6},
			{"version", 0, 0, 7},
			{"help", 0, 0, 8},
			{"output", 0, 0, 9},
			{0, 0, 0, 0}
		};

		c = getopt_long( argc, argv, "i:o:", cdinput_options, &option_index );

		if( c == -1 )
			break;
		switch( c )
		{
			case 'i':
			case 0:
				if( strcmp( optarg, SHORT_NAME ) != 0 )
					stop_reading_args = 1;
				break;
			case 1:
				strncpy( cdinput_args->mountpoint, optarg, PATH_MAX );
				cdinput_args->Options |= CDINPUT_ARGS_MOUNTPOINT;
				break;
			case 2:
				strncpy( cdinput_args->cddevice, optarg, PATH_MAX );
				cdinput_args->Options |= CDINPUT_ARGS_DEVICE;
				break;
			case 3:
				strncpy( cdinput_args->filesystem, optarg, 255 );
				cdinput_args->Options |= CDINPUT_ARGS_FILESYSTEM;
				break;
			case 4:
				strncpy( cdinput_args->cdlabel, optarg, 33 );
				cdinput_args->Options |= CDINPUT_ARGS_CDLABEL;
				break;
			case 5:
				cdinput_args->Options |= CDINPUT_ARGS_AUTOMOUNT;
				break;
			case 6:
				cdinput_args->Options |= CDINPUT_ARGS_EJECT;
				break;
			case 7:
				cdinput_args->Options |= CDINPUT_ARGS_VERSION;
				break;
			case 8:
				cdinput_args->Options |= CDINPUT_ARGS_HELP;
				break;
			case 'o':
			case 9:
					stop_reading_args = 1;
			default:
		}

		if( stop_reading_args == 1 )
			break;
	}

	return 0;
}


int
check_args( int argc, char *argv[] )
{
	struct CDInputArgs cdinput_args;
	int ok_args = 0;

	ok_args = read_args( argc, argv, &cdinput_args );


	/**
	 * If the user typo the read_args function returns
	 * -1. This will print out an usage prompt to the user
	 */
	if( ok_args == -1 )
		return -1;

	/* Do the user want the version? */
	if( cdinput_args.Options & CDINPUT_ARGS_VERSION )
	{
		version();
		return -1;
	}

	/* Do the user want usage info? */
	if( cdinput_args.Options & CDINPUT_ARGS_HELP )
	{
		usage();
		return -1;
	}

	/* Check for required arguments */
	if( strlen( cdinput_args.mountpoint ) == 0 )
	{
		usage();
		return -1;
	}

	return 0;
}

static PluginStruct CDInputStructure = 
{
	SHORT_NAME,
	"CD Input",
	"Input from CD-Rom",
	1,
	NULL,
	read_cd,
	NULL,
	check_args,
	NULL,
	NULL,
	NULL
};

PluginStruct *
load_plugin()
{

	return &CDInputStructure;
}
