/*
 * Copyright (C) 2001 the MCatS project
 *
 * This file is part of the MCatS, Modular Catalog System.
 *
 * MCatS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MCatS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <getopt.h>
#include <malloc.h>
#include "mcats.h"

#define SHORT_NAME "iodbc"
#define IODBC_ARGS_HELP 1
#define IODBC_ARGS_DISKNUMBER 2
#define IODBC_ARGS_DSN 4
#define IODBC_ARGS_CONNECTSTRING 8
#define IODBC_ARGS_VOLUMETABLE 16
#define IODBC_ARGS_DIRECTORYTABLE 32
#define IODBC_ARGS_FILETABLE 64
#define IODBC_ARGS_VOLUMEFORMAT 128
#define IODBC_ARGS_DIRECTORYFORMAT 256
#define IODBC_ARGS_FILEFORMAT 512
#define IODBC_ARGS_VERSION 1024

#define IODBC_OPTARGS_DISKLEFT	1

struct IODBCArgs {
	int disknumber;
	char dsn[33];
	char connectstring[129];
	char volumetable[255];
	char directorytable[255];
	char filetable[255];
	char volumeformat[255];
	char directoryformat[255];
	char fileformat[255];
	int Options;
};

struct IODBCOptArgs {
	int output_diskleft;
	int Options;
};

void input_usage()
{
	printf( "usage: mcats -i iodbc [options]\n"
			"options:\n"
			"   --disknumber [disknumber]Witch fields to print to stdout\n"
			"   --help                   Prints this help and exit (optional)\n\n");
}

void output_usage()
{
	printf( "usage: mcats -o iodbc [options]\n"
			"options:\n"
			"   --dsn\n"
			"   --connectstring\n"
			"   --volumetable\n"
			"   --directorytable\n"
			"   --filetable\n"
			"   --volumeformat\n"
			"   --directoryformat\n"
			"   --fileformat\n"
			"   --help                   Prints this help and exit (optional)\n\n");
}

void version()
{
	printf( "iodbc version %s\n", IODBC_VERSION );
}

/* Take the supplied input argument diskleft and pick out the next
 * one to be read. Also rewrite the input argument diskleft so
 * the one that will be read is removed from the list.
 */
int
parsedisknumber( struct MainStruct *mainstructure, int argc_index )
{
	char *tmp_argv;
	char s_disknumber[5] = "";
	int disknumber = 0;

	if( strlen( mainstructure->input_argv[argc_index-1] ) == 0 )
	{
		// No arguments left. Cleanup and quit
		free( mainstructure->input_argv[argc_index-1] );
		free( mainstructure->input_argv[argc_index-2] );
		free( mainstructure->input_argv );
		mainstructure->input_argc = -1;
		return -1;
	}

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: argc_index = !%u!\n", argc_index );
#endif

	tmp_argv = strdup( strchr( mainstructure->input_argv[argc_index-1], ',' ) + 1 );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: tmp_argv = !%s!\n", tmp_argv );
#endif

	strncpy( s_disknumber, mainstructure->input_argv[argc_index-1], strlen( mainstructure->input_argv[argc_index-1] ) - strlen( tmp_argv ) - 1 );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: s_disknumber = !%s!\n", s_disknumber );
#endif

	disknumber = atoi( s_disknumber );
	strcpy( mainstructure->input_argv[argc_index-1], tmp_argv );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: input_argv = !%s!\n", mainstructure->input_argv[argc_index-1] );
#endif

	free( tmp_argv );
	return disknumber;
}

struct VolumeStruct
read_iodbc ( struct MainStruct *mainstructure )
{
	struct IODBCArgs iodbc_args;
	struct IODBCOptArgs iodbc_optargs;
	struct VolumeStruct volumestructure;
	int disknumber = 0;
	int ok_args = 0;
	int result = 0;
	int loop = 0;

	/* set filestructure to NULL */
	volumestructure.filestructure = NULL;

	ok_args = read_input_args( mainstructure->argc, mainstructure->argv, &iodbc_args );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: input_argc = %d\n", mainstructure->input_argc );
	if( mainstructure->input_argv != NULL )
		fprintf( stderr, "DEBUG: input_argv != NULL\n" );
#endif
	if( mainstructure->input_argc != -1 && mainstructure->input_argv != NULL )
		ok_args = read_opt_args( mainstructure->input_argc, mainstructure->input_argv, &iodbc_optargs );



#ifdef __DEBUG__
	fprintf( stderr, "Looking for disk: %u\n", disknumber );
#endif

	return volumestructure;
}

int
write_iodbc ( struct MainStruct *mainstructure, struct VolumeStruct volumestructure )
{
	struct IODBCArgs iodbc_args;
	char db_connectstring[163] = "";
	int ok_args = 0;
	int id = 0;

	ok_args = read_output_args( mainstructure->argc, mainstructure->argv, &iodbc_args );

	strcpy( db_connectstring, iodbc_args.dsn );
	strcat( db_connectstring, ";" );
	strcat( db_connectstring, iodbc_args.connectstring );
	if( DB_Connect( db_connectstring ) != 0 )
	{
		/* Couldn't connect to ODBC source */
		fprintf( stderr, "ERROR: Couldn't connect to data source: %s\n", db_connectstring );
		return -1;
	}

	if( ( iodbc_args.Options & IODBC_ARGS_VOLUMETABLE )
		&& ( iodbc_args.Options & IODBC_ARGS_VOLUMEFORMAT ) )
	id = DB_InsertVolume( volumestructure, iodbc_args.volumetable, iodbc_args.volumeformat );

	if( ( iodbc_args.Options & IODBC_ARGS_DIRECTORYTABLE )
		&& ( iodbc_args.Options & IODBC_ARGS_DIRECTORYFORMAT )
		&& ( iodbc_args.Options & IODBC_ARGS_FILETABLE )
		&& ( iodbc_args.Options & IODBC_ARGS_FILEFORMAT ) )
	DB_InsertFileStruct( volumestructure.filestructure, iodbc_args.directorytable, iodbc_args.filetable, iodbc_args.directoryformat, iodbc_args.fileformat, id );

	DB_Disconnect();

	return 0;
}


int
read_input_args( int argc, char *argv[], struct IODBCArgs *iodbc_args )
{
	int c;
	int stop_reading_args = 0;
	int loop = 0;
	/* Set getopt's index pointer */
	while( strcmp( argv[loop++], "-i" ) != 0 )
		;
	optind = loop - 1;

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: argv[optind+2] = !%s!\n", argv[optind+2] );
#endif

/* Clear the structure so we can se if it was filled */
	iodbc_args->disknumber = 0;
	iodbc_args->Options = 0;


	while(1)
	{
		int option_index = 0;

		static struct option iodbc_options[] =
		{
			{"listdsn", 0, 0, 0},
			{"dsn", 1, 0, 1},
			{"connectstring", 1, 0, 2},
			{"volumetable", 1, 0, 3},
			{"disknumber", 1, 0, 4},
			{"help", 0, 0, 5},
			{0, 0, 0, 0}
		};

		c = getopt_long( argc, argv, "o:i:", iodbc_options, &option_index );

		if( c == -1 )
			break;

		switch( c )
		{
			case 'i':
				if( strcmp( optarg, SHORT_NAME ) == 0 )
					stop_reading_args = 0;
				else
					stop_reading_args = 1;
				break;
			case 0:
				DB_ListDSN();
				return 1;
				break;
			case 1:
				iodbc_args->Options |= IODBC_ARGS_DSN;
				break;
			case 2:
				iodbc_args->Options |= IODBC_ARGS_CONNECTSTRING;
				break;
			case 3:
				iodbc_args->Options |= IODBC_ARGS_VOLUMETABLE;
				break;
			case 4:
				iodbc_args->disknumber = atoi( optarg );
				iodbc_args->Options |= IODBC_ARGS_DISKNUMBER;
				break;
			case 5:
				iodbc_args->Options |= IODBC_ARGS_HELP;
				break;
			case 'o':
				stop_reading_args = 1;
				break;
			default:
				break;
		}

		if( stop_reading_args == 1 )
			break;
	}

	return 0;
}

int
read_output_args( int argc, char *argv[], struct IODBCArgs *iodbc_args )
{
	int c;
	int stop_reading_args = 0;
	int loop = 0;
	/* Set getopt's index pointer */
	while( strcmp( argv[loop++], "-o" ) != 0 )
		;
	optind = loop - 1;


	/* Clear the structure so we can se if it was filled */
	iodbc_args->Options = 0;
	strcpy( iodbc_args->dsn, "" );
	strcpy( iodbc_args->connectstring, "" );
	strcpy( iodbc_args->volumetable, "" );
	strcpy( iodbc_args->directorytable, "" );
	strcpy( iodbc_args->filetable, "" );
	strcpy( iodbc_args->volumeformat, "" );
	strcpy( iodbc_args->directoryformat, "" );
	strcpy( iodbc_args->fileformat, "" );


	while(1)
	{
		int option_index = 0;

		static struct option iodbc_options[] =
		{
			{"output", 1, 0, 0},
			{"listdsn", 0, 0, 1},
			{"dsn", 1, 0, 2},
			{"connectstring", 1, 0, 3},
			{"volumetable", 1, 0, 4},
			{"directorytable", 1, 0, 5},
			{"filetable", 1, 0, 6},
			{"volumeformat", 1, 0, 7},
			{"directoryformat", 1, 0, 8},
			{"fileformat", 1, 0, 9},
			{"help", 0, 0, 10},
			{"input", 1, 0, 11},
			{0, 0, 0, 0}
		};

		c = getopt_long( argc, argv, "o:i:", iodbc_options, &option_index );

		if( c == -1 )
			break;

		switch( c )
		{
			case 'o':
			case 0:
				if( strcmp( optarg, SHORT_NAME ) == 0 )
					stop_reading_args = 0;
				else
					stop_reading_args = 1;
				break;
			case 1:
				DB_ListDSN();
				return 1;
				break;
			case 2:
				strcpy( iodbc_args->dsn, "DSN=" );
				strncat( iodbc_args->dsn, optarg, 28 );
				iodbc_args->Options |= IODBC_ARGS_DSN;
				break;
			case 3:
				strncpy( iodbc_args->connectstring, optarg, 128 );
				iodbc_args->Options |= IODBC_ARGS_CONNECTSTRING;
				break;
			case 4:
				strncpy( iodbc_args->volumetable, optarg, 254 );
				iodbc_args->Options |= IODBC_ARGS_VOLUMETABLE;
				break;
			case 5:
				strncpy( iodbc_args->directorytable, optarg, 254 );
				iodbc_args->Options |= IODBC_ARGS_DIRECTORYTABLE;
				break;
			case 6:
				strncpy( iodbc_args->filetable, optarg, 254 );
				iodbc_args->Options |= IODBC_ARGS_FILETABLE;
				break;
			case 7:
				strncpy( iodbc_args->volumeformat, optarg, 254 );
				iodbc_args->Options |= IODBC_ARGS_VOLUMEFORMAT;
				break;
			case 8:
				strncpy( iodbc_args->directoryformat, optarg, 254 );
				iodbc_args->Options |= IODBC_ARGS_DIRECTORYFORMAT;
				break;
			case 9:
				strncpy( iodbc_args->fileformat, optarg, 254 );
				iodbc_args->Options |= IODBC_ARGS_FILEFORMAT;
				break;
			case 10:
				iodbc_args->Options |= IODBC_ARGS_HELP;
				break;
			case 'i':
			case 11:
				stop_reading_args = 1;
				break;
			default:
				break;
		}

	}

	return 0;
}

int
read_opt_args( int argc, char *argv[], struct IODBCOptArgs *iodbc_optargs )
{
	int c;
	int stop_reading_args = 0;
	int loop = 0;
	/* Set getopt's index pointer */
	while( strcmp( argv[loop++], "--output_diskleft" ) != 0 )
		;
	optind = loop - 1;

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: optind = %d\n", optind );
	fprintf( stderr, "DEBUG: argc = %d\n", argc );
	fprintf( stderr, "DEBUG: argv[optind] = !%s!\n", argv[optind] );
	fprintf( stderr, "DEBUG: argv[optind+1] = !%s!\n", argv[optind+1] );
#endif

/* Clear the structure so we can se if it was filled */
	iodbc_optargs->output_diskleft = 0;
	iodbc_optargs->Options = 0;


	while(1)
	{
		int option_index = 0;

		static struct option iodbc_opt_options[] =
		{
			{"output_diskleft", 1, 0, 0},
			{0, 0, 0, 0}
		};

		c = getopt_long( argc, argv, "", iodbc_opt_options, &option_index );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: c = %d\n", c );
#endif

		if( c == -1 )
			break;

		switch( c )
		{
			case 0:
				iodbc_optargs->output_diskleft = optind;
				iodbc_optargs->Options |= IODBC_OPTARGS_DISKLEFT;
#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: Got --output_diskleft!!\n" );
#endif
				break;
			default:
				break;
		}
	}

	return 0;
}

int
check_input_args( int argc, char *argv[] )
{
	struct IODBCArgs iodbc_args;
	int ok_args = 0;

	ok_args = read_input_args( argc, argv, &iodbc_args );

	/**
	 * If the user typo the read_args function returns
	 * -1. This will print out an usage prompt to the user
	 */
	if( ok_args == -1 )
		return -1;

	/*
	 * If the command was already executed in read_input_args
	 * return 1 so it quits normaly
	 */
	if( ok_args == 1 )
		return 1;


	/* Check for required arguments */
	// TODO

	/* Do the user want usage info? */
	if( iodbc_args.Options & IODBC_ARGS_HELP )
	{
		input_usage();
		return -1;
	}

	return 0;
}

int
check_output_args( int argc, char *argv[] )
{
	struct IODBCArgs iodbc_args;
	int ok_args = 0;

	ok_args = read_output_args( argc, argv, &iodbc_args );

	/**
	 * If the user typo the read_args function returns
	 * -1. This will print out an usage prompt to the user
	 */
	if( ok_args == -1 )
		return -1;

	/*
	 * If the command was already executed in read_input_args
	 * return 1 so it quits normaly
	 */
	if( ok_args == 1 )
		return 1;

	/* Do the user want the version? */
	if( iodbc_args.Options & IODBC_ARGS_VERSION )
	{
		version();
		return -1;
	}

	/* Do the user want usage info? */
	if( iodbc_args.Options & IODBC_ARGS_HELP )
	{
		output_usage();
		return -1;
	}

	/* Check for required arguments */
	if( ! ( iodbc_args.Options & IODBC_ARGS_DSN ) )
		return -1;

	/* Enter in case the user put in the right combination of arguments */
	if( ( iodbc_args.Options & IODBC_ARGS_VOLUMETABLE ) != 0
			&& ( iodbc_args.Options & IODBC_ARGS_VOLUMEFORMAT ) != 0
			&& ( iodbc_args.Options & IODBC_ARGS_DIRECTORYTABLE ) == 0
			&& ( iodbc_args.Options & IODBC_ARGS_DIRECTORYFORMAT ) == 0
			&& ( iodbc_args.Options & IODBC_ARGS_FILETABLE ) == 0
			&& ( iodbc_args.Options & IODBC_ARGS_FILEFORMAT ) == 0 )
	{
		return 0;
	}
	else if( ( iodbc_args.Options & IODBC_ARGS_VOLUMETABLE ) == 0
			&& ( iodbc_args.Options & IODBC_ARGS_VOLUMEFORMAT ) == 0
			&& ( iodbc_args.Options & IODBC_ARGS_DIRECTORYTABLE ) != 0
			&& ( iodbc_args.Options & IODBC_ARGS_DIRECTORYFORMAT ) != 0
			&& ( iodbc_args.Options & IODBC_ARGS_FILETABLE ) != 0
			&& ( iodbc_args.Options & IODBC_ARGS_FILEFORMAT ) != 0 )
	{
		return 0;
	}
	else if( ( iodbc_args.Options & IODBC_ARGS_VOLUMETABLE ) != 0
			&& ( iodbc_args.Options & IODBC_ARGS_VOLUMEFORMAT ) != 0
			&& ( iodbc_args.Options & IODBC_ARGS_DIRECTORYTABLE ) != 0
			&& ( iodbc_args.Options & IODBC_ARGS_DIRECTORYFORMAT ) != 0
			&& ( iodbc_args.Options & IODBC_ARGS_FILETABLE ) != 0
			&& ( iodbc_args.Options & IODBC_ARGS_FILEFORMAT ) != 0 )
	{
		return 0;
	}

	/* If the user missed one argument it comes here */
	output_usage();
	return -1;

}

static PluginStruct IODBCStructure = 
{
	SHORT_NAME,
	"IODBC",
	"Read/Write to ODBC source",
	2,
	write_iodbc,
	read_iodbc,
	NULL,
	check_input_args,
	check_output_args,
	NULL,
	NULL
};

PluginStruct *
load_plugin()
{
	return &IODBCStructure;
}
