/*
* Copyright (C) 2001 the MCatS project
*
* This file is part of the MCatS, Modular Catalog System.
*
* MCatS is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* MCatS is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
*
*/

#include <stdio.h>
#include <string.h>
#include <sql.h>
#include <sqlext.h>

#include "mcats.h"

#define MAXCOLS		32

SQLHENV henv;
SQLHDBC hdbc;
SQLHSTMT hstmt;
int connected;

char *requestcpy( char *, char * );
char *requestncpy( char *, char *, int );
char *requestcat( char *, char * );
char *requestncat( char *, char *, int );

char *requestrcat( char *, char * );
char *requestrncat( char *, char *, int );
char *requestrcpy( char *, char * );
char *requestrncpy( char *, char *, int );

/*
* This is original the odbctest.c distributed with libiodbc.
* It has been heavily rewritten but still want to note where its orig from.
*/

/*
* Connect to the datasource
*
* The connect string can have the following parts and they refer to
* the values in the odbc.ini file
*
* DSN=<data source name>		[mandatory]
* HOST=<server host name>		[optional - value of Host]
* SVT=<database server type>		[optional - value of ServerType]
* DATABASE=<database path>		[optional - value of Database]
* OPTIONS=<db specific opts>		[optional - value of Options]
* UID=<user name>			[optional - value of LastUser]
* PWD=<password>			[optional]
* READONLY=<N|Y>			[optional - value of ReadOnly]
* FBS=<fetch buffer size>		[optional - value of FetchBufferSize]
*
* Examples:
*
* HOST=star;SVT=Informix 5;UID=demo;PWD=demo;DATABASE=stores5
*
* DSN=stores5_informix;PWD=demo
*/
int
DB_Connect (char *connStr)
{
	short buflen;
	char buf[257];
	SQLCHAR dataSource[120];
	SQLCHAR dsn[33];
	SQLCHAR desc[255];
	SWORD len1, len2;
	int status;

#if (ODBCVER < 0x0300)
	if (SQLAllocEnv (&henv) != SQL_SUCCESS)
		return -1;

	if (SQLAllocConnect (henv, &hdbc) != SQL_SUCCESS)
		return -1;
#else
	if (SQLAllocHandle (SQL_HANDLE_ENV, NULL, &henv) != SQL_SUCCESS)
		return -1;

	SQLSetEnvAttr (henv, SQL_ATTR_ODBC_VERSION, (SQLPOINTER) SQL_OV_ODBC3, SQL_IS_UINTEGER);

	if (SQLAllocHandle (SQL_HANDLE_DBC, henv, &hdbc) != SQL_SUCCESS)
		return -1;
#endif

	strcpy ((char *) dataSource, connStr);

	status = SQLDriverConnect (hdbc, 0, (UCHAR *) dataSource, SQL_NTS,
			(UCHAR *) buf, sizeof (buf), &buflen, SQL_DRIVER_COMPLETE);

	if (status != SQL_SUCCESS && status != SQL_SUCCESS_WITH_INFO)
		return -1;

	// SQLSetConnectOption (hdbc, SQL_OPT_TRACEFILE, (UDWORD) "\\SQL.LOG");

	connected = 1;

#if (ODBCVER < 0x0300)
	if (SQLAllocStmt (hdbc, &hstmt) != SQL_SUCCESS)
		return -1;
#else
	if (SQLAllocHandle (SQL_HANDLE_STMT, hdbc, &hstmt) != SQL_SUCCESS)
		return -1;
#endif

	return 0;
}


/*
* Disconnect from the database
*/
int
DB_Disconnect (void)
{
#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: Doing DB_Disconnect()\n" );
#endif

#if (ODBCVER < 0x0300)
	if (hstmt)
		SQLFreeStmt (hstmt, SQL_DROP);

	if (connected)
		SQLDisconnect (hdbc);

	if (hdbc)
		SQLFreeConnect (hdbc);

	if (henv)
		SQLFreeEnv (henv);
#else
	if (hstmt)
		SQLFreeHandle (SQL_HANDLE_STMT, hstmt);

	if (connected)
		SQLDisconnect (hdbc);

	if (hdbc)
		SQLFreeHandle (SQL_HANDLE_DBC, hdbc);

	if (henv)
		SQLFreeHandle (SQL_HANDLE_ENV, henv);
#endif

	return 0;
}


/*
* This is the message handler for the communications layer.
*
* The messages received here are not passed through SQLError,
* because they might occur when no connection is established.
*
* Typically, Rejections from oplrqb are trapped here, and
* also RPC errors.
*
* When no message handler is installed, the messages are output to stderr
*/
void
DB_MesgHandler (char *reason)
{
	fprintf (stderr, "DB_MesgHandler: %s\n", reason);
}


/*
* Show all the error information that is available
*/
int
DB_Errors (char *where)
{
	unsigned char buf[250];
	unsigned char sqlstate[15];

	/*
	 * Get statement errors
	 */
	while (SQLError (henv, hdbc, hstmt, sqlstate, NULL,
			buf, sizeof(buf), NULL) == SQL_SUCCESS)
	{
		fprintf (stderr, "%s, SQLSTATE=%s\n", buf, sqlstate);
	}

	/*
	 * Get connection errors
	 */
	while (SQLError (henv, hdbc, SQL_NULL_HSTMT, sqlstate, NULL,
			buf, sizeof(buf), NULL) == SQL_SUCCESS)
	{
		fprintf (stderr, "%s, SQLSTATE=%s\n", buf, sqlstate);
	}

	/*
	 * Get environmental errors
	 */
	while (SQLError (henv, SQL_NULL_HDBC, SQL_NULL_HSTMT, sqlstate, NULL,
			buf, sizeof(buf), NULL) == SQL_SUCCESS)
	{
		fprintf (stderr, "%s, SQLSTATE=%s\n", buf, sqlstate);
	}

	return -1;
}


/*
*	
*/
int
DB_InsertVolume ( struct VolumeStruct volumestructure, char *table, char *volumeformat )
{
	char *request;
	char fetchBuffer[1000];
	short colWidths[MAXCOLS];
	short numCols;
	short colNum;
	char colName[SQL_MAX_COLUMN_NAME_LEN];
	short colType;
	UDWORD colPrecision;
	SDWORD colIndicator;
	short colScale;
	short colNullable;
	UDWORD totalRows;
	int i, sts;
	int id_value = -1;
	int id = -1;
	int loop = 0;
	char ch = 0;
	char id_name[SQL_MAX_COLUMN_NAME_LEN] = "";
	char s_tmp[255] = "";
	/* For format string */
	char **content;
	char *toc;
	int fields = 0;
	int max_field_length = 0;

#ifdef __DEBUG__
	fprintf( stderr, "Entering DB_InsertVolume()\n" );
#endif
	/*
	 * Parse the format string so we know which column the ID is in.
	 */
#ifdef __DEBUG__
	fprintf( stderr, "Parsing format string\n" );
#endif
	fields = getFieldCount( volumeformat, &max_field_length );

	toc = (char *) malloc( fields );
	content = (char **)calloc( fields, max_field_length );

	id = parseFormatString( volumeformat, content, toc, fields, max_field_length );

	/*
	 * Start by getting the last id entry so we can pick out which
	 * entry we just inserted incase there is identical entries.
	 * We also get an array that tells if the active column is a
	 * value or a string.
	 */
	request = requestcpy( request, "SELECT * FROM " );
	request = requestcat( request, table );
	request = requestcat( request, " ORDER BY " );
	sprintf( s_tmp, "%u DESC;", id );
	request = requestcat( request, s_tmp );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: Executing \"%s\"\n", request );
#endif

	if (SQLPrepare (hstmt, (UCHAR *) request, SQL_NTS) != SQL_SUCCESS)
	{
		DB_Errors ("SQLPrepare");
		goto endCursor;
	}
	if (SQLExecute (hstmt) != SQL_SUCCESS)
	{
		DB_Errors ("SQLExec");
		goto endCursor;
	}

	if (SQLNumResultCols (hstmt, &numCols) != SQL_SUCCESS)
	{
		DB_Errors ("SQLNumResultCols");
		goto endCursor;
	}

	if (numCols == 0)
	{
		fprintf ( stderr, "numCols=0\n");
		goto endCursor;
	}

	if (numCols > MAXCOLS)
		numCols = MAXCOLS;

	/* Create an array of the column types */
	for (colNum = 1; colNum <= numCols; colNum++)
	{
		if (SQLDescribeCol (hstmt, colNum, (UCHAR *) colName,
					sizeof (colName), NULL, &colType, &colPrecision,
					&colScale, &colNullable) != SQL_SUCCESS)
		{
			DB_Errors ("SQLDescribeCol");
			goto endCursor;
		}

		/*
		 * Get the name of the ID column so we can retrieve information
		 * from it later.
		 */
		if( colNum == id )
			strncpy( id_name, colName, SQL_MAX_COLUMN_NAME_LEN );

		switch (colType)
		{
			case SQL_VARCHAR:
			case SQL_CHAR:
				colWidths[colNum-1] = (short) colPrecision;
				break;
			case SQL_BIT:
			case SQL_TINYINT:
			case SQL_SMALLINT:
			case SQL_INTEGER:
			case SQL_DOUBLE:
			case SQL_DECIMAL:
			case SQL_NUMERIC:
			case SQL_FLOAT:
				colWidths[colNum-1] = -1;
				break;
			case SQL_DATE:
			case SQL_TIME:
			case SQL_TIMESTAMP:
				/*
				 * Presume it's a string since it's more likely to get a date
				 * from the volumestructure then the user.
				 */
		  	colWidths[colNum-1] = (short) -colPrecision;
				break;
		default:
				/* Other datatypes will be taken as value */
				colWidths[colNum-1] = -1;
				break;
		}
	}


	sts = SQLFetch (hstmt);

	if (sts == SQL_NO_DATA_FOUND)
	{
		/* If this is the case, it will be the only one avaible after insert so id_value = 0 then */
		id_value = 0;
		goto insertion;
	}

	if (sts != SQL_SUCCESS)
	{
		DB_Errors ("Fetch");
		goto endCursor;
	}

	/*
	 *	Fetch this column as character
	 */
	if (SQLGetData (hstmt, 1, SQL_CHAR, fetchBuffer,
				sizeof (fetchBuffer), &colIndicator) != SQL_SUCCESS)
	{
		DB_Errors ("SQLGetData");
		goto endCursor;
	}

	id_value = atoi( fetchBuffer );

insertion:
	SQLFreeStmt (hstmt, SQL_CLOSE);
	/* Insert volume */
	request = requestcpy( request, "INSERT INTO " );
	request = requestcat( request, table );
	request = requestcat( request, " VALUES ( " );
	/* This is where the parser should put together the values in the right order */
#ifdef __DEBUG__
	fprintf( stderr, "volumeformat: !%s!\n", volumeformat );
	fprintf( stderr, "strlen(toc): %u\n", strlen( toc ) );
#endif



	/* Build the SQL query acording to the user input */
	colNum = 0;
	if( colWidths[colNum] >= 0 || ( colWidths[colNum] < -1 && toc[1] == 'c' ) )
		request = requestcat( request, "'" );
	for( loop = 0; loop < strlen( toc ); loop++ )
	{
		if( toc[loop] == 'c' )
			ch = content[loop][0];
		else if( toc[loop] == 'e' )
			ch = toupper( content[loop][0] );
		else if( toc[loop] == 's' )
			ch = 0;

		switch( ch ) {
			case 'a':
				/* Add the volumename to SQL query */
				request = requestrcat( request, volumestructure.name );
				break;
			case 'b':
				/* Add the publisher to SQL query */
				request = requestrcat( request, volumestructure.publisher );
				break;
			case 'c':
				/* Add the preparer to SQL query */
				request = requestrcat( request, volumestructure.preparer );
				break;
			case 'd':
				/* Add the application to SQL query */
				request = requestrcat( request, volumestructure.application );
				break;
			case 'e':
				/* Add the type of filesystem to SQL query */
				sprintf( s_tmp, "%ld", volumestructure.f_type );
				request = requestcat( request, s_tmp );
				break;
			case 'f':
				/* Add copy of mount flags to SQL query (deprecated) */
				sprintf( s_tmp, "%ld", volumestructure.f_oflags );
				request = requestcat( request, s_tmp );
				break;
			case 'g':
				/* Add fundamental file system block size to SQL query */
				sprintf( request, "%l", volumestructure.f_bsize );
				request = requestcat( request, s_tmp );
				break;
			case 'h':
				/* Add optimal transfer block size to SQL query */
				sprintf( s_tmp, "%l", volumestructure.f_iosize );
				request = requestcat( request, s_tmp );
				break;
			case 'i':
				/* Add total data blocks in file system to SQL query */
				sprintf( s_tmp, "%l", volumestructure.f_blocks );
				request = requestcat( request, s_tmp );
				break;
			case 'j':
				/* Add free blocks in fs to SQL query */
				sprintf( s_tmp, "%l", volumestructure.f_bfree );
				request = requestcat( request, s_tmp );
				break;
			case 'k':
				/* Add free blocks avail to non-superuser to SQL query */
				sprintf( s_tmp, "%l", volumestructure.f_bavail );
				request = requestcat( request, s_tmp );
				break;
			case 'l':
				/* Add total file nodes in file system to SQL query */
				sprintf( s_tmp, "%l", volumestructure.f_files );
				request = requestcat( request, s_tmp );
				break;
			case 'm':
				/* Add free file nodes in fs to SQL query */
				sprintf( s_tmp, "%l", volumestructure.f_ffree );
				request = requestcat( request, s_tmp );
				break;
			case 'n':
				/* Add file system id to SQL query */
				sprintf( s_tmp, "%l", volumestructure.f_fsid );
				request = requestcat( request, s_tmp );
				break;
			case 'o':
				/* Add user that mounted the file system to SQL query */
				sprintf( s_tmp, "%l", volumestructure.f_owner );
				request = requestcat( request, s_tmp );
				break;
			case 'p':
				/* Add copy of mount flags to SQL query */
				sprintf( s_tmp, "%l", volumestructure.f_flags );
				request = requestcat( request, s_tmp );
				break;
			case 'q':
				/* Add count of sync writes since mount to SQL query */
				sprintf( s_tmp, "%l", volumestructure.f_syncwrites );
				request = requestcat( request, s_tmp );
				break;
			case 'r':
				/* Add count of async writes since mount to SQL query */
				sprintf( s_tmp, "%l", volumestructure.f_asyncwrites );
				request = requestcat( request, s_tmp );
				break;
			case 't':
				/* Add fs type name to SQL query */
				request = requestrcat( request, volumestructure.f_fstypename );
				break;
			case 'u':
				/* Add directory on which mounted to SQL query */
				request = requestrcat( request, volumestructure.f_mntonname );
				break;
			case 'v':
				/* Add mounted file system to SQL query */
				request = requestrcat( request, volumestructure.f_mntfromname );
				break;
			case 0:
				if( strcmp( content[loop], "*" ) == 0 )
					request = requestcat( request, "''" );
				else
					request = requestcat( request, content[loop] );
				break;
			case '|':
				/* Add column seperator to SQL query */
				if( colWidths[colNum] >= 0 || ( colWidths[colNum] < -1 && toc[loop+1] == 'c' ) )
					request = requestcat( request, "'" );

				colNum++;
				request = requestcat( request, "," );

				if( colWidths[colNum] >= 0 || ( colWidths[colNum] < -1 && toc[loop+1] == 'c' ) )
					request = requestcat( request, "'" );
				break;
			case 'N':
				/* Add a newline to SQL query */
				request = requestcat( request, "\n" );
				break;
			case 'T':
				/* Add a horizontal tab to SQL query */
				request = requestcat( request, "\t" );
				break;
			case '\\':
				/* Add a backslash to SQL query */
				request = requestcat( request, "\\" );
				break;
			default:
				/* Add as string to SQL query */
				request = requestcat( request, content[loop] );
				break;
		}
	}
	/* Clear memory allocated for format string */
	for( loop = 0; loop < fields; loop++)
		free( content[loop] );

	free(content);
	free(toc);

	if( colWidths[colNum] >= 0 )
		request = requestcat( request, "'" );
	request = requestcat( request, " );" );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: Executing \"%s\"\n", request );
#endif
	/*
	 * Prepare & Execute the statement
	 */
	if (SQLPrepare (hstmt, (UCHAR *) request, SQL_NTS) != SQL_SUCCESS)
	{
		DB_Errors ("SQLPrepare");
		goto endCursor;
	}
	if (SQLExecute (hstmt) != SQL_SUCCESS)
	{
		DB_Errors ("SQLExec");
		goto endCursor;
	}

	/*
	 * Get the number of result columns for this cursor.
	 * If it is 0, then the statement was probably not a select
	 */
	if (SQLNumResultCols (hstmt, &numCols) != SQL_SUCCESS)
	{
		DB_Errors ("SQLNumResultCols");
		goto endCursor;
	}
	if (numCols == 0)
	{
		/* This should happend since we do a insert */
#ifdef __DEBUG__
		fprintf (stderr, "Statement executed.\n");
#endif
	} 
		
	/* Get the id the volume got */
	request = requestcpy( request, "SELECT * FROM " );
	request = requestcat( request, table );
	request = requestcat( request, " WHERE " );
	request = requestcat( request, id_name );
	request = requestcat( request, " >" );
	sprintf( s_tmp, "%d", id_value );
	request = requestcat( request, s_tmp );
	request = requestcat( request, " ORDER BY " );
	request = requestcat( request, id_name );
	request = requestcat( request, " DESC;" );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: Executing \"%s\"\n", request );
#endif

	if (SQLPrepare (hstmt, (UCHAR *) request, SQL_NTS) != SQL_SUCCESS)
	{
		DB_Errors ("SQLPrepare");
		goto endCursor;
	}
	if (SQLExecute (hstmt) != SQL_SUCCESS)
	{
		DB_Errors ("SQLExec");
		goto endCursor;
	}
	if (numCols > MAXCOLS)
		numCols = MAXCOLS;

	/*
	 * Print all the fields
	 */
	sts = SQLFetch (hstmt);

	if (sts == SQL_NO_DATA_FOUND)
		goto endCursor;

	if (sts != SQL_SUCCESS)
	{
		DB_Errors ("Fetch");
		goto endCursor;
	}

	/*
	 * Fetch this column as character
	 */
	if (SQLGetData (hstmt, 1, SQL_CHAR, fetchBuffer,
				sizeof (fetchBuffer), &colIndicator) != SQL_SUCCESS)
	{
		DB_Errors ("SQLGetData");
		goto endCursor;
	}

	/*
	 * Show NULL fields as ****
	 */
	if (colIndicator == SQL_NULL_DATA)
	{
		fetchBuffer[0] = -1;
		fetchBuffer[1] = '\0';
	}

	id_value = atoi( fetchBuffer );

endCursor:
	SQLFreeStmt (hstmt, SQL_CLOSE);

#ifdef __DEBUG__
	fprintf( stderr, "Leaving DB_InsertVolume()\n" );
#endif
	return id_value;
}

/*
*	
*/
int
DB_InsertFileStruct ( struct FileStruct *first_file_in_struct, char *dirtable, char *filetable, char *directoryformat, char *fileformat, int volumeid )
{
	struct FileStruct *current_file_in_struct;
	char *request;
	char fetchBuffer[1000];
	short colWidths[MAXCOLS];
	short numCols;
	short colNum;
	char colName[50];
	short colType;
	UDWORD colPrecision;
	SDWORD colIndicator;
	short colScale;
	short colNullable;
	UDWORD totalRows;
	int i, sts;
	char *table;
	char *format;
	int id = -1;
	int loop = 0;
	char ch;
	char s_tmp[256] = "";
	/* For format string */
	char **content;
	char *toc;
	int fields = 0;
	int max_field_length = 0;

#ifdef __DEBUG__
	fprintf( stderr, "Entering DB_InsertFileStruct()\n" );
#endif
	/*
	 * Parse the format string so we know which column the ID is in.
	 */
#ifdef __DEBUG__
	fprintf( stderr, "Parsing format string\n" );
#endif
	fields = getFieldCount( fileformat, &max_field_length );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: max_field_length=%u\n", max_field_length );
#endif
	toc = (char *) malloc( fields );
	content = (char **)calloc( fields, max_field_length );

	id = parseFormatString( fileformat, content, toc, fields, max_field_length );


	/*
	 * Start by getting the last id entry so we can pick out which
	 * entry we just inserted incase there is identical entries
	 */
	request = requestcpy( request, "SELECT * FROM " );
	request = requestcat( request, dirtable );
	request = requestcat( request, " ORDER BY " ); 
	sprintf( s_tmp, "%u DESC;", id );
	request = requestcat( request, s_tmp );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: Executing \"%s\"\n", request );
#endif

	if( SQLPrepare (hstmt, (UCHAR *) request, SQL_NTS) != SQL_SUCCESS )
	{
		DB_Errors ("SQLPrepare");
		goto endCursor;
	}
	if (SQLExecute (hstmt) != SQL_SUCCESS)
	{
		DB_Errors ("SQLExec");
		goto endCursor;
	}

	if (SQLNumResultCols (hstmt, &numCols) != SQL_SUCCESS)
	{
		DB_Errors ("SQLNumResultCols");
		goto endCursor;
	}
	if (numCols == 0)
	{
		fprintf ( stderr, "numCols=0\n");
		goto endCursor;
	}
		
	if (numCols > MAXCOLS)
		numCols = MAXCOLS;

	/* Create an array of the column types */
	for (colNum = 1; colNum <= numCols; colNum++)
	{
		if (SQLDescribeCol (hstmt, colNum, (UCHAR *) colName,
					sizeof (colName), NULL, &colType, &colPrecision,
					&colScale, &colNullable) != SQL_SUCCESS)
		{
			DB_Errors ("SQLDescribeCol");
			goto endCursor;
		}

		switch (colType)
		{
			case SQL_VARCHAR:
			case SQL_CHAR:
				colWidths[colNum-1] = (short) colPrecision;
				break;
			case SQL_BIT:
			case SQL_TINYINT:
			case SQL_SMALLINT:
			case SQL_INTEGER:
			case SQL_DOUBLE:
			case SQL_DECIMAL:
			case SQL_NUMERIC:
			case SQL_FLOAT:
				colWidths[colNum-1] = -1;
				break;
			case SQL_DATE:
			case SQL_TIME:
			case SQL_TIMESTAMP:
				/*
				 * Presume it's a string since it's more likely to get a date
				 * from the volumestructure then the user.
				 */
		  	colWidths[colNum-1] = (short) -colPrecision;
				break;
		default:
				/* Other datatypes will be taken as value */
				colWidths[colNum-1] = -1;
				break;
		}
	}

	sts = SQLFetch (hstmt);

	if (sts == SQL_NO_DATA_FOUND)
	{
		/* If this is the case, it will be the only one avaible after insert so id = 0 then */
		id = 0;
		goto insertion;
	}

	if (sts != SQL_SUCCESS)
	{
		DB_Errors ("Fetch");
		goto endCursor;
	}

	/*
	 *	Fetch this column as character
	 */
	if (SQLGetData (hstmt, 1, SQL_CHAR, fetchBuffer, sizeof (fetchBuffer), &colIndicator) != SQL_SUCCESS)
	{
		DB_Errors ("SQLGetData");
		goto endCursor;
	}

insertion:
	SQLFreeStmt (hstmt, SQL_CLOSE);

	current_file_in_struct = first_file_in_struct;
	while( current_file_in_struct->next != 0 )
	{
		if( S_ISDIR( current_file_in_struct->mode ) )
		{
			table = dirtable;
			format = directoryformat;
		}
		else
		{
			table = filetable;
			format = fileformat;
		}

		/* Insert volume */
		request = requestcpy( request, "INSERT INTO " );
		request = requestcat( request, table );
		request = requestcat( request, " VALUES ( " );
		/* This is where the parser should put together the values in the right order */
#ifdef __DEBUG__
		fprintf( stderr, "directoryformat: !%s!\n", directoryformat );
#endif


		/* Build the SQL query acording to the user input */
		colNum = 0;
		if( colWidths[colNum] >= 0 || ( colWidths[colNum] < -1 && toc[1] == 'c' ) )
			request = requestcat( request, "'" );
		for( loop = 0; loop < strlen( toc ); loop++ )
		{
			if( toc[loop] == 'c' )
				ch = content[loop][0];
			else if( toc[loop] == 'e' )
				ch = toupper( content[loop][0] );
			else if( toc[loop] == 's' )
				ch = 0;

			switch( ch ) {
				case 'f':
					/* Add the filename to SQL query */
					request = requestrcat( request, current_file_in_struct->filename );
					break;
				case 'p':
					/* Add the path to SQL query */
					request = requestrcat( request, current_file_in_struct->path );
					break;
				case 's':
					/* Add the size to SQL query */
					sprintf( s_tmp, "%ld", current_file_in_struct->size );
					request = requestcat( request, s_tmp );
					break;
				case 't':
					/* Add the type to SQL query */
					sprintf( s_tmp, "%d", current_file_in_struct->type );
					request = requestcat( request, s_tmp );
					break;
				case 'm':
					/* Add the mode to SQL query */
					sprintf( s_tmp, "%u", current_file_in_struct->mode );
					request = requestcat( request, s_tmp );
					break;
				case 'o':
					/* Add the owner to SQL query */
					sprintf( s_tmp, "%u", current_file_in_struct->owner );
					request = requestcat( request, s_tmp );
					break;
				case 'g':
					/* Add the group to SQL query */
					sprintf( s_tmp, "%u", current_file_in_struct->group );
					request = requestcat( request, s_tmp );
					break;
				case 'i':
					/* Add the mime to SQL query */
					request = requestrcat( request, current_file_in_struct->mime );
					break;
				case 'd':
					/* Add the description to SQL query */
					request = requestcat( request, "'" );
					if( current_file_in_struct->description != NULL )
						request = requestrcat( request, current_file_in_struct->description );
					request = requestcat( request, "'" );
				case 0:
					if( strcmp( content[loop], "*" ) == 0 )
						request = requestcat( request, "''" );
					else
						request = requestcat( request, content[loop] );
					break;
				case '|':
					/* Add column seperator to SQL query */
					if( colWidths[colNum] >= 0 || ( colWidths[colNum] < -1 && toc[loop+1] == 'c' ) )
						request = requestcat( request, "'" );

					colNum++;
					request = requestcat( request, "," );

					if( colWidths[colNum] >= 0 || ( colWidths[colNum] < -1 && toc[loop+1] == 'c' ) )
						request = requestcat( request, "'" );
					break;
				case 'V':
					sprintf( s_tmp, "%d", volumeid );
					request = requestcat( request, s_tmp );
					break;
			}
		}


		if( colWidths[colNum] >= 0 )
			request = requestcat( request, "'" );
		request = requestcat( request, " );" );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: Executing \"%s\"\n", request );
#endif
		/*
		 * Prepare & Execute the statement
		 */
		if (SQLPrepare (hstmt, (UCHAR *) request, SQL_NTS) != SQL_SUCCESS)
		{
			DB_Errors ("SQLPrepare");
			goto endCursor;
		}
		if (SQLExecute (hstmt) != SQL_SUCCESS)
		{
			DB_Errors ("SQLExec");
			goto endCursor;
		}

		/*
		 * Get the number of result columns for this cursor.
		 * If it is 0, then the statement was probably not a select
		 */
		if (SQLNumResultCols (hstmt, &numCols) != SQL_SUCCESS)
		{
			DB_Errors ("SQLNumResultCols");
			goto endCursor;
		}
		if (numCols == 0)
		{
			/* This should happend since we do a insert */
#ifdef __DEBUG__
			printf ("Statement executed.\n");
#endif
		} 
		
		SQLFreeStmt (hstmt, SQL_CLOSE);
		current_file_in_struct = current_file_in_struct->next;
	}
endCursor:
	SQLFreeStmt (hstmt, SQL_CLOSE);

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: freeing %u fields from content\n", fields );
#endif
	/* Clear memory allocated for format string */
	for( loop = 0; loop < fields; loop++)
		free( content[loop] );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: freeing content\n" );
#endif
	free(content);
#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: freeing toc\n" );
#endif
	free(toc);

#ifdef __DEBUG__
	fprintf( stderr, "Leaving DB_InsertFileStruct()\n" );
#endif
	return id;
}

int
DB_ListDSN()
{
	SQLCHAR dsn[33];
	SQLCHAR desc[255];
	SWORD len1, len2;

#if (ODBCVER < 0x0300)
	if (SQLAllocEnv (&henv) != SQL_SUCCESS)
		return -1;

	if (SQLAllocConnect (henv, &hdbc) != SQL_SUCCESS)
		return -1;
#else
	if (SQLAllocHandle (SQL_HANDLE_ENV, NULL, &henv) != SQL_SUCCESS)
		return -1;

	SQLSetEnvAttr (henv, SQL_ATTR_ODBC_VERSION, (SQLPOINTER) SQL_OV_ODBC3,
			SQL_IS_UINTEGER);

	if (SQLAllocHandle (SQL_HANDLE_DBC, henv, &hdbc) != SQL_SUCCESS)
		return -1;
#endif

	/*
	 * Print headers
	 */
	printf ( "\n%-33s | %-44s\n", "DSN", "Description" );
	printf ( "--------------------------------------------------------------------------------\n" );

	/*
	 * Goto the first record
	 */
	if (SQLDataSources (henv, SQL_FETCH_FIRST,
				dsn, 33, &len1, desc, 255, &len2) != SQL_SUCCESS)
		return -1;

	/*
	 * Show all records
	 */
	do
	{
		fprintf (stderr, "%-33s | %-44s\n", dsn, desc);
	} while (SQLDataSources (henv, SQL_FETCH_NEXT,
				dsn, 33, &len1, desc, 255, &len2) == SQL_SUCCESS);

	DB_Disconnect();
}

int
getFieldCount( char *formatstring, int *max_field_length )
{
	char *stringp;
	char *seps;
	char nextchar = 0;              /* Next character in inputstring */
	int fields = 0;

	stringp = strdup( formatstring );
	while( seps = strsep( &stringp, "%\\|" ) )
	{
		if( stringp != 0 )
		{
			nextchar = *(stringp++);
			fields++;
		}
		else
			nextchar = 0;
		if( strlen( seps ) > (*max_field_length) )
			(*max_field_length) = strlen( seps );
		if( strlen( seps ) != 0 )
			fields++;
	}

	/* It seems that max_field_length must be larger than it should be */
	(*max_field_length) += 2;

	return fields;
}


/*
* Parse the format string and set the content and toc aproparate.
* Return the column where the ID is located ( first column is 1 ).
*/
int
parseFormatString( char *formatstring, char **content, char *toc, int fields, int max_field_length )
{
	/* For the stringparser */
	char *stringp;                          /* Pointer to original input string */
	char *seps;                                      /* Pointer to seperated string */
	char prevchar = 0;              /* Previous character in inputstring */
	char nextchar = 0;              /* Next character in inputstring */
	int previndex = 0;
	int loop = 0;
	int ch = 0;                                      /* Tmp var to store current character */
	int index = 0;                          /* Index when processing the inputstring */
	int id_index = 1;
	int id_field = 0;


	for( loop = 0; loop < fields; loop++ )
		content[loop] = (char *) malloc( max_field_length + 1 );

	stringp = strdup( formatstring );
	while( seps = strsep( &stringp, "%\\|" ) )
	{
		if( stringp != 0 )
		{
			previndex = strlen( formatstring ) - strlen( stringp ) - 1;
			/* Get the previous character, %, \ or | */
			prevchar = formatstring[previndex];
			nextchar = *(stringp++);
		}
		else
			prevchar = nextchar = 0;

		if( strlen( seps ) != 0 )
		{
			toc[index++] = 's';
			sprintf( content[index - 1], "%s", seps );
			if( strcmp( content[index - 1], "*" ) == 0 && toc[index - 1] == 'e' && nextchar == '|' )
				id_field = id_index;
		}
		if( nextchar != 0 )
		{
			if( prevchar == '%' )
			{
				toc[index++] = 'c';
				sprintf( content[index - 1], "%c", nextchar );
			} else if( prevchar == '\\' )
			{
				toc[index++] = 'e';
				sprintf( content[index - 1], "%c", nextchar );
			} else if( prevchar == '|' )
			{
				id_index++;
				toc[index++] = 'e';
				sprintf( content[index - 1], "%c", prevchar );
				stringp--;
			}
		}
	}
	toc[index] = '\0';

	return id_field+1;
}

char *
requestcpy( char *dest, char *src )
{
	return requestncpy( dest, src, strlen( src ) );
}

char *
requestncpy( char *dest, char *src, int len )
{
	dest = (char *)malloc( len + 1 );

	/* If there was any problem allocating new memory */
	if( dest == NULL )
		return dest;

	strncpy( dest, src, len );
	dest[len] = 0;

	return dest;
}

char *
requestcat( char *dest, char *src )
{
	return requestncat( dest, src, strlen( src ) );
}

char *
requestncat( char *dest, char *src, int len )
{
	char *newdest;
	int oldlength = strlen(dest);

	newdest = (char *)realloc( dest, strlen( dest ) + len + 1 );

	/* If there was any problem allocating new memory */
	if( newdest == NULL )
	{
		fprintf( stderr, "ERROR: Couldn't allocate memory for SQL Query!\n" );
		return dest;
	}

	dest = newdest;
	strncat( dest, src, len );
	dest[oldlength+len] = 0;

	return dest;
}

char *
requestrcat( char *dest, char *src )
{
	return requestrncat( dest, src, strlen(src) );
}

char *
requestrncat( char *dest, char *src, int len )
{
	char *newdest;
	int srcloop = 0, destloop = 0;

	newdest = (char *)malloc( len * 2 + 1 );

	/* If there was any problem allocating new memory */
	if( newdest == NULL )
		return dest;

	for( srcloop = 0, destloop = 0; srcloop < len; srcloop++ )
	{
		if( src[srcloop] == '\''
			|| src[srcloop] == '\"'
			|| src[srcloop] == '\\' )
		{
			newdest[destloop++] = '\\';
			newdest[destloop++] = src[srcloop];
		}
		else
			newdest[destloop++] = src[srcloop];
	}
	newdest[destloop] = 0;

	dest = requestcat( dest, newdest );

	return dest;
}

char *
requestrcpy( char *dest, char *src )
{
	return requestrncat( dest, src, strlen(src) );
}

char *
requestrncpy( char *dest, char *src, int len )
{
	char *newdest;
	int srcloop = 0, destloop = 0;

	newdest = (char *)malloc( len * 2 + 1 );

	/* If there was any problem allocating new memory */
	if( newdest == NULL )
		return dest;

	for( srcloop = 0, destloop = 0; srcloop < len; srcloop++ )
	{
		if( src[srcloop] == '\''
			|| src[srcloop] == '\"'
			|| src[srcloop] == '\\' )
		{
			newdest[destloop++] = '\\';
			newdest[destloop++] = src[srcloop];
		}
		else
			newdest[destloop++] = src[srcloop];
	}
	newdest[destloop] = 0;

	dest = requestcpy( dest, newdest );

	return dest;
}

/*
 * Know problems:
 */
