/*
 * Copyright (C) 2001 the MCatS project
 *
 * This file is part of the MCatS, Modular Catalog System.
 *
 * MCatS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MCatS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>
#include <getopt.h>
#include <malloc.h>
#include "mcats.h"
#include "wii.h"

#define SHORT_NAME "wii"
#define WII_ARGS_HELP 1
#define WII_ARGS_INPUTFILE 2
#define WII_ARGS_DISKNUMBER	8
#define WII_ARGS_VERSION 16

#define WII_OPTARGS_DISKLEFT	1

struct WIIArgs {
	char inputfile[256];
	int disknumber;
	int Options;
};

struct WIIOptArgs {
	int input_diskleft;
	int Options;
};

void input_usage()
{
	printf( "usage: mcats -i wii [options]\n"
		"options:\n"
		"   --inputfile [filename]   WhereIsIt file to read from\n"
		"   --disknumber [number]    Wich disk to read\n"
		"   --version                Prints wii version and exit\n"
		"   --help                   Prints this help and exit\n\n");
}

void version()
{
	printf( "wii version %s\n", WII_VERSION );
}

struct VolumeStruct
wii300_readinputfile( FILE *fp, int disknumber, struct MainStruct *mainstructure, int argc_index )
{
	struct VolumeStruct volumestructure;
	struct FileStruct *first_file_in_struct;
	struct FileStruct *current_file_in_struct;
	struct FileStruct *new_file_in_struct;
	struct FileStruct *prev_file_in_struct;
	struct FileStruct *tmp_file_in_struct;
	struct wii300header *header;
	struct wii300file *filestruct;
	struct wii300dir *dirstruct;
	long file_startpos = 0;
	int filesindirs = 0;
	int dirlevel = 1;
	int result = 0;
	int loop = 0;
	int loop2 = 0;
	char tmppath[255] = "";

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: Doing wii300_readinputfile()\n" );
	printf( "errno: !%u!\n", errno );
#endif

	/* Read the header */
	header = wii300_readtopheader( fp );

#ifdef __DEBUG__
	if( disknumber == -1 )
	{
		fprintf( stderr, "DEBUG: argc_index = %u\n", argc_index );
		fprintf( stderr, "DEBUG: input_argc = %u\n", mainstructure->input_argc );
		fprintf( stderr, "DEBUG: strlen( input_argv ) = %u\n", strlen( mainstructure->input_argv[argc_index-1] ) );
	}
#endif
	if( disknumber == -1 && mainstructure->input_argv != NULL )
		if( strlen( mainstructure->input_argv[argc_index-1] ) == 0 )
		{
			// set disknumber to the first disk
			disknumber = header->disknumbers[0];
			for( loop = 1; loop < header->totaldisks; loop++ )
			{
				sprintf( mainstructure->input_argv[argc_index-1], "%s%u,", mainstructure->input_argv[argc_index-1], header->disknumbers[loop] );
			}
		}

#ifdef __DEBUG__
	if( disknumber == -1 )
		fprintf( stderr, "DEBUG: input_argv = !%s!\n", mainstructure->input_argv[argc_index-1] );
#endif

	/* Search to given disknumber */
	result = wii300_finddisk( fp, header, disknumber );

#ifdef __DEBUG__
	fprintf( stderr, "Result %d, when searching for disk: %u\n", result, disknumber );
#endif

	if( result != 0 )
	{
		fprintf( stderr, "Disknumber %d wasn't found\n", disknumber );
		volumestructure.filestructure = NULL;
		return volumestructure;
	}

	wii300_readheader( fp, header );

#ifdef __DEBUG__
	fprintf( stderr, "header->totaldisks: %u\n", header->totaldisks );
	fprintf( stderr, "header->disknumbers: " );
	for( loop = 0; loop < header->totaldisks; loop++ )
		fprintf( stderr, "%u,", header->disknumbers[loop] );
	fprintf( stderr, "\n" );
	fprintf( stderr, "header->catalogname: !%s!\n", header->catalogname );
	fprintf( stderr, "header->diskname: !%s!\n", header->diskname );
	fprintf( stderr, "header->format: %u\n", header->format );
	fprintf( stderr, "header->disknumber: %u\n", header->disknumber );
	fprintf( stderr, "header->bytespersector: %u\n", header->bytespersector );
	fprintf( stderr, "header->sectorspercluster: %u\n", header->sectorspercluster );
	fprintf( stderr, "header->maxfilelength: %u\n", header->maxfilelength );
	fprintf( stderr, "header->totalfiles: %u\n", header->totalfiles );
	fprintf( stderr, "header->totaldirs: %u\n", header->totaldirs );
	fprintf( stderr, "header->totalclusters: %u\n", header->totalclusters );
	fprintf( stderr, "header->freeclusters: %u\n", header->freeclusters );
	fprintf( stderr, "header->serialnumber: %c%c%c%c\n", header->serialnumber[0], header->serialnumber[1], header->serialnumber[2], header->serialnumber[3] );
	fprintf( stderr, "header->medianame: !%s!\n", header->medianame );
	fprintf( stderr, "header->fsname: !%s!\n", header->fsname );
#endif

	/* Fill the volumestructure with the values we got */
	strcpy( volumestructure.name, header->diskname );
	strcpy( volumestructure.f_fstypename, header->fsname );
	volumestructure.f_type = 0;
	volumestructure.f_oflags = 0;
	volumestructure.f_bsize = header->bytespersector * header->sectorspercluster;
	volumestructure.f_iosize = 0;
	volumestructure.f_blocks = header->totalclusters;
	volumestructure.f_bfree = header->freeclusters;
	volumestructure.f_bavail = header->freeclusters;
	volumestructure.f_files = 0;
	volumestructure.f_ffree = 0;
	volumestructure.f_owner = 0;
	volumestructure.f_flags = 0;
	volumestructure.f_syncwrites = 0;
	volumestructure.f_asyncwrites = 0;
	volumestructure.f_fstypename[0] = 0;
	volumestructure.f_mntonname[0] = 0;
	volumestructure.f_mntfromname[0] = 0;


	file_startpos = ftell( fp );

	/* Search forward until the directories starts */
	for( loop = 0; loop < header->totalfiles; loop++ )
	{
		filestruct = wii300_readfile( fp );
		wii300_freefile( filestruct );
	}

	/*
	 * Read directories and store them in the structure.
	 * Also makes room for the files to be put in inbetween
	 */
	first_file_in_struct = (struct FileStruct *) malloc( sizeof( struct FileStruct ) );

	current_file_in_struct = first_file_in_struct;
	current_file_in_struct->description = NULL;
	for( loop = 0; loop < header->totaldirs; loop++ )
	{
		dirstruct = wii300_readdir( fp );

		/* If it's the first dir, use / as path */
		if( dirstruct->placeintree == 1 )
			sprintf( current_file_in_struct->path, "/" );
		else if( dirstruct->placeintree == dirlevel )
			sprintf( current_file_in_struct->path, "%s", prev_file_in_struct->path );
		else if( dirstruct->placeintree > dirlevel )
			sprintf( current_file_in_struct->path, "%s%s/",
							prev_file_in_struct->path, prev_file_in_struct->filename );
		else if( dirstruct->placeintree < dirlevel )
		{
			memset( current_file_in_struct->path, '\0', sizeof( current_file_in_struct->path ) );
			/* Now we have to split up so we get to the right level.. */
			loop2 = dirstruct->placeintree;
			/* Get ride of the last slash so we can see how many chars the last dir is */
			strncpy( tmppath, prev_file_in_struct->path, strlen(prev_file_in_struct->path) - 1);
			do {
				tmppath[strlen(tmppath)-strlen(strrchr(tmppath, '/'))] = '\0';
			} while( ++loop2 < dirlevel );
			strcpy( current_file_in_struct->path, tmppath );
			strcat( current_file_in_struct->path, "/" );
		}
		/* Put in the dir in the filename field (Yes, this is correct) */
		sprintf( current_file_in_struct->filename, "%s", dirstruct->dirname );

		/* Set directory mode */
		current_file_in_struct->mode = S_IFDIR;

		/* Put in the creation date into modification date..
		 * dont know if that's the best idea.. but hell, it's beta.. :)
		 */
		current_file_in_struct->st_ctimespec = dirstruct->timestamp;

#ifdef __DEBUG__
	printf( "dirstruct->files: %u\n", dirstruct->files );
#endif
		filesindirs += dirstruct->files;

		/* Set the dirlevel used this time so we got something to compare too */
		dirlevel = dirstruct->placeintree;
		prev_file_in_struct = current_file_in_struct;
		for( loop2 = 0; loop2 <= dirstruct->files; loop2++ )
		{

#ifdef __DEBUG__
	fprintf( stderr, "allocate memory for file\n" );
	fprintf( stderr, "errno: !%u!\n", errno );
#endif

			new_file_in_struct = ( struct FileStruct * )
														malloc( sizeof( struct FileStruct ) );

			current_file_in_struct->next = new_file_in_struct;
			current_file_in_struct = new_file_in_struct;
			sprintf( current_file_in_struct->path, "%s%s/", prev_file_in_struct->path, prev_file_in_struct->filename  );
			current_file_in_struct->st_ctimespec = 0;
			/* Set directory mode */
			current_file_in_struct->mode = 0;
			current_file_in_struct->description = NULL;
		}
		wii300_freedir( dirstruct );
	}

	/*
	 * Allocate memory for the files in the root.
	 */

	tmp_file_in_struct = first_file_in_struct;
	if( filesindirs < header->totalfiles )
	{
		first_file_in_struct = ( struct FileStruct * ) malloc( sizeof( struct FileStruct ) );
		strcpy( first_file_in_struct->path, "/" );
		first_file_in_struct->st_ctimespec = 0;
		first_file_in_struct->description = NULL;
		current_file_in_struct = first_file_in_struct;
	}
	for( loop = filesindirs; loop < header->totalfiles-1; loop++ )
	{
		new_file_in_struct = ( struct FileStruct * )
													malloc( sizeof( struct FileStruct ) );

		current_file_in_struct->next = new_file_in_struct;
		current_file_in_struct = new_file_in_struct;
		strcpy( current_file_in_struct->path, "/" );
		current_file_in_struct->st_ctimespec = 0;
		current_file_in_struct->description = NULL;
	}
	if( filesindirs < header->totalfiles )
		current_file_in_struct->next = tmp_file_in_struct;

#ifdef __DEBUG__
	printf( "charafterdirs: %u\n", fgetc( fp ) );
#endif

	fseek( fp, file_startpos, SEEK_SET );

	/* Put in the files into the structures */
	current_file_in_struct = first_file_in_struct;
	for( loop = 0; loop < header->totalfiles; loop++ )
	{
		filestruct = wii300_readfile( fp );
		while( current_file_in_struct->st_ctimespec != 0 )
		{
			current_file_in_struct = current_file_in_struct->next;
		}
		sprintf( current_file_in_struct->filename, "%s", filestruct->filename );
		current_file_in_struct->size = filestruct->filesize;
		current_file_in_struct->st_ctimespec = filestruct->timestamp;
#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: desc: %u\n", filestruct->desc );
#endif
		if( filestruct->desc == 2 )
		{
			current_file_in_struct->description = malloc( strlen( filestruct->filedescription ) + 1 );
			strncpy( current_file_in_struct->description, filestruct->filedescription, strlen( filestruct->filedescription ) );
		}

		current_file_in_struct = current_file_in_struct->next;
		//current_file_in_struct->description = NULL;

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: wii.current_file_in_struct: %u\n", current_file_in_struct );
	//if( current_file_in_struct->description != NULL )
	//	fprintf( stderr, "DEBUG: wii.current_file_in_struct->description: %u\n", current_file_in_struct->description );
#endif

		wii300_freefile( filestruct );
	}

	/* Set current_file_in_struct->next to NULL just to be sure */
	current_file_in_struct->next = NULL;

	wii300_freeheader( header );

	fclose( fp );

	volumestructure.filestructure = first_file_in_struct;

#ifdef __DEBUG__
	printf( "Leaving function\n" );
#endif

	return volumestructure;
}

/* Take the supplied input argument diskleft and pick out the next
 * one to be read. Also rewrite the input argument diskleft so
 * the one that will be read is removed from the list.
 */
int
parsedisknumber( struct MainStruct *mainstructure, int argc_index )
{
	char *tmp_argv;
	char s_disknumber[5] = "";
	int disknumber = -1;

	if( strlen( mainstructure->input_argv[argc_index-1] ) == 0 )
	{
		// No arguments left. Cleanup and quit
		free( mainstructure->input_argv[argc_index-1] );
		free( mainstructure->input_argv[argc_index-2] );
		free( mainstructure->input_argv );
		mainstructure->input_argc = -1;
		return -2;
	}

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: argc_index = !%u!\n", argc_index );
#endif

	tmp_argv = strdup( strchr( mainstructure->input_argv[argc_index-1], ',' ) + 1 );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: tmp_argv = !%s!\n", tmp_argv );
#endif

	strncpy( s_disknumber, mainstructure->input_argv[argc_index-1], strlen( mainstructure->input_argv[argc_index-1] ) - strlen( tmp_argv ) - 1 );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: s_disknumber = !%s!\n", s_disknumber );
#endif

	disknumber = atoi( s_disknumber );
	strcpy( mainstructure->input_argv[argc_index-1], tmp_argv );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: input_argv = !%s!\n", mainstructure->input_argv[argc_index-1] );
#endif

	free( tmp_argv );
	return disknumber;
}
struct VolumeStruct
read_wii ( struct MainStruct *mainstructure )
{
	FILE *fp;
	struct WIIArgs wii_args;
	struct WIIOptArgs wii_optargs;
	struct VolumeStruct volumestructure;
	int disknumber = -1;
	int ok_args = 0;
	int loop = 0;


	ok_args = read_input_args( mainstructure->argc, mainstructure->argv, &wii_args );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: input_argc = %d\n", mainstructure->input_argc );
	if( mainstructure->input_argv != NULL )
		fprintf( stderr, "DEBUG: input_argv != NULL\n" );
#endif
	if( mainstructure->input_argc != -1 && mainstructure->input_argv != NULL )
		ok_args = read_opt_args( mainstructure->input_argc, mainstructure->input_argv, &wii_optargs );
	else
		wii_optargs.Options = 0;

	if( wii_args.Options & WII_ARGS_INPUTFILE )
	{
		if( ( fp = fopen( wii_args.inputfile, "rb" ) ) == NULL )
		{
			fprintf( stderr, "ERROR: '%s' No such file or directory\n", wii_args.inputfile );
			volumestructure.filestructure = NULL;
			return volumestructure;
		}

		if( wii_args.disknumber >= 0 )
			disknumber = wii_args.disknumber;

		if( !( wii_args.Options & WII_ARGS_DISKNUMBER ) &&
				!( wii_optargs.Options & WII_OPTARGS_DISKLEFT ) )
		{
			// More? less? Calc it...
			if( mainstructure->input_argc == -1 )
			{
#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: input_argc is -1\n" );
#endif
				if( mainstructure->input_argv == NULL )
				{
					mainstructure->input_argv = malloc( 20 * sizeof( mainstructure->input_argv ) );
#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: input_argv is NULL\n" );
#endif
				}
				mainstructure->input_argc = 0;
				mainstructure->input_argv[++mainstructure->input_argc] = malloc( 255 );
				strncpy( mainstructure->input_argv[mainstructure->input_argc], "--input_diskleft", 254 );
				mainstructure->input_argv[++mainstructure->input_argc] = malloc( 4096 );
				memset( mainstructure->input_argv[mainstructure->input_argc], '\0', 4096 );
#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: allocated space and set input_argv[0] to !%s!\n", mainstructure->input_argv[mainstructure->input_argc-1] );
#endif
			}
			else
			{
#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: NO --input_diskleft allocating new space\n" );
#endif
				mainstructure->input_argc = 0;
				mainstructure->input_argv[++mainstructure->input_argc] = malloc( 255 );
				strncpy( mainstructure->input_argv[mainstructure->input_argc], "--input_diskleft", 254 );
				mainstructure->input_argv[++mainstructure->input_argc] = malloc( 4096 );
			}
			wii_optargs.input_diskleft = ++mainstructure->input_argc;
		}

	if( wii_optargs.Options & WII_OPTARGS_DISKLEFT )
	{
		disknumber = parsedisknumber( mainstructure, wii_optargs.input_diskleft );
		if( disknumber == -2 )
		{
			volumestructure.filestructure = NULL;
			return volumestructure;
		}
	}


#ifdef __DEBUG__
	fprintf( stderr, "Looking for disk: %d\n", disknumber );
#endif

		if( check300version( fp ) == 0 )
			volumestructure = wii300_readinputfile( fp, disknumber, mainstructure, wii_optargs.input_diskleft );
		else {
			/* Not right version on the file, quitting */
			fprintf( stderr, "ERROR: %s is not in WhereIsIt 3.00 format\n", wii_args.inputfile );
			volumestructure.filestructure = NULL;
			return volumestructure;
		}
	}

	return volumestructure;
}


int
read_input_args( int argc, char *argv[], struct WIIArgs *wii_args )
{
	int c;
	int stop_reading_args = 0;
	int loop = 0;
	/* Set getopt's index pointer */
	while( strcmp( argv[loop++], "-i" ) != 0 )
		;
	optind = loop - 1;

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: argv[optind+2] = !%s!\n", argv[optind+2] );
#endif

/* Clear the structure so we can se if it was filled */
	sprintf( wii_args->inputfile, "" );
	wii_args->disknumber = -1;
	wii_args->Options = 0;


	while(1)
	{
		int option_index = 0;

		static struct option wii_options[] =
		{
			{"input", 1, 0, 0},
			{"inputfile", 1, 0, 1},
			{"disknumber", 1, 0, 2},
			{"help", 0, 0, 3},
			{"output", 1, 0, 4},
			{0, 0, 0, 0}
		};

		c = getopt_long( argc, argv, "o:i:", wii_options, &option_index );

		if( c == -1 )
			break;

		switch( c )
		{
			case 'i':
			case 0:
				if( strcmp( optarg, SHORT_NAME ) == 0 )
					stop_reading_args = 0;
				else
					stop_reading_args = 1;
				break;
			case 1:
				strncpy( wii_args->inputfile, optarg, 255 );
				wii_args->Options |= WII_ARGS_INPUTFILE;
				break;
			case 2:
				wii_args->disknumber = atoi( optarg );
				wii_args->Options |= WII_ARGS_DISKNUMBER;
				break;
			case 3:
				wii_args->Options |= WII_ARGS_HELP;
				break;
			case 'o':
			case 4:
					stop_reading_args = 1;
					break;
			default:
				break;
		}

		if( stop_reading_args == 1 )
			break;
	}

	return 0;
}

int
read_opt_args( int argc, char *argv[], struct WIIOptArgs *wii_optargs )
{
	int c;
	int stop_reading_args = 0;
	int loop = 1;
	/* Set getopt's index pointer */
	while( strcmp( argv[loop++], "--input_diskleft" ) != 0 )
		;
	optind = loop - 1;

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: optind = %d\n", optind );
	fprintf( stderr, "DEBUG: argc = %d\n", argc );
	fprintf( stderr, "DEBUG: argv[optind] = !%s!\n", argv[optind] );
	fprintf( stderr, "DEBUG: argv[optind+1] = !%s!\n", argv[optind+1] );
#endif

/* Clear the structure so we can se if it was filled */
	wii_optargs->input_diskleft = 0;
	wii_optargs->Options = 0;


	while(1)
	{
		int option_index = 0;

		static struct option wii_opt_options[] =
		{
			{"input_diskleft", 1, 0, 0},
			{0, 0, 0, 0}
		};

		c = getopt_long( argc, argv, "", wii_opt_options, &option_index );

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: c = %d\n", c );
#endif

		if( c == -1 )
			break;

		switch( c )
		{
			case 0:
				wii_optargs->input_diskleft = optind;
				wii_optargs->Options |= WII_OPTARGS_DISKLEFT;
#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: Got --input_diskleft!!\n" );
#endif
				break;
			default:
				break;
		}
	}

	return 0;
}

int
check_input_args( int argc, char *argv[] )
{
	struct WIIArgs wii_args;
	int ok_args = 0;

	ok_args = read_input_args( argc, argv, &wii_args );

	/**
	 * If the user typo the read_args function returns
	 * -1. This will print out an usage prompt to the user
	 */
	if( ok_args == -1 )
		return -1;

	/* Do the user want the version? */
	if( wii_args.Options & WII_ARGS_VERSION )
	{
		usage();
		return -1;
	}

	/* Do the user want usage info? */
	if( wii_args.Options & WII_ARGS_HELP )
	{
		input_usage();
		return -1;
	}

	/* Check for required arguments */
	if( ! ( wii_args.Options & WII_ARGS_INPUTFILE ) )
	{
		input_usage();
		return -1;
	}

	return 0;
}

static PluginStruct WIIStructure = 
{
	SHORT_NAME,
	"WhereIsIt",
	"Input from WhereIsIt file",
	1,
	NULL,
	read_wii,
	NULL,
	check_input_args,
	NULL,
	NULL,
	NULL
};

PluginStruct *
load_plugin()
{
	return &WIIStructure;
}
