/*
 * Copyright (C) 2001 the MCatS project
 *
 * This file is part of the MCatS, Modular Catalog System.
 *
 * MCatS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MCatS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
 *
 */

#include <stdio.h>
#include <time.h>
#include <malloc.h>
#include "wii.h"


#define IDENTSTRING "Catalog 3.00"

int
check300version( FILE *fp )
{
	char ident[13] = "";

	if( fread( ident, 12, 1, fp ) < 1 )
	{
		perror("ERROR");
		return -1;
	}

	if( strcmp( ident, IDENTSTRING ) == 0 )
		return 0;
	else
		return -1;
}

int
wii300_finddisk( FILE *fp, struct wii300header *header, int disknumber )
{
	struct wii300file *filestruct;
	struct wii300dir *dirstruct;
	long diskstartpos = 0;
	int loop = 0;

	do
	{

#ifdef __DEBUG__
	fprintf( stderr, "ERROR: feof, %u\n", feof(fp) );
#endif

		diskstartpos = ftell( fp );

		wii300_readheader( fp, header );

		if( feof( fp ) != 0 )
			return -1;

#ifdef __DEBUG__
  printf( "header->totaldisks: %u\n", header->totaldisks );
  printf( "header->type: %u\n", header->type );
  printf( "header->disknumbers: " );
  for( loop = 0; loop < header->totaldisks; loop++ )
    printf( "%u,", header->disknumbers[loop] );
  printf( "\n" );
  printf( "header->catalogname: !%s!\n", header->catalogname );
  printf( "header->diskname: !%s!\n", header->diskname );
  printf( "header->format: %u\n", header->format );
  printf( "header->disknumber: %u\n", header->disknumber );
  printf( "header->bytespersector: %u\n", header->bytespersector );
  printf( "header->sectorspercluster: %u\n", header->sectorspercluster );
  printf( "header->maxfilelength: %u\n", header->maxfilelength );
  printf( "header->totalfiles: %u\n", header->totalfiles );
  printf( "header->totaldirs: %u\n", header->totaldirs );
  printf( "header->totalclusters: %u\n", header->totalclusters );
  printf( "header->freeclusters: %u\n", header->freeclusters );
  printf( "header->medianame: !%s!\n", header->medianame );
  printf( "header->fsname: !%s!\n", header->fsname );
#endif

		/* Search forward until the directories starts */
		for( loop = 0; loop < header->totalfiles; loop++ )
		{
			filestruct = wii300_readfile( fp );
			wii300_freefile( filestruct );
		}

		/* Search forward until the start of new header */
		for( loop = 0; loop < header->totaldirs; loop++ )
		{
			dirstruct = wii300_readdir( fp );
			wii300_freedir( dirstruct );
		}

#ifdef __DEBUG__
	fprintf( stderr, "DEBUG: Found disk: %u, looking for: %u\n", header->disknumber, disknumber );
#endif

	} while( header->disknumber != disknumber && feof(fp) == 0 );

	if( feof( fp ) != 0 )
		return -1;

	/* Set the fp pos to where the header for the disk starts */
	if( diskstartpos != 0 )
		fseek( fp, diskstartpos, SEEK_SET );

	return 0;
}

struct wii300header *
wii300_readtopheader( FILE *fp )
{
	struct wii300header *header;
	unsigned char charcount = 0;

	// Allocate some memory for the header
	header = (struct wii300header *) malloc( sizeof( struct wii300header ) );

	wii300_readtotaldisks( fp, header );

	wii300_readdisknumbers( fp, header );

	wii300_readcatalogname( fp, header );

	return header;
}

void
wii300_readheader( FILE *fp, struct wii300header *header )
{
	wii300_readheadertype( fp, header );
	wii300_readdiskname( fp, header );

// Should be read in a seperate section
// Because this is probably the header every following disk will have

	wii300_readheaderformat( fp, header );

	// If the header isn't for iso9660
	if( header->format != 522 && header->format != 778 )
		return;

	wii300_readdisknumber( fp, header );

	wii300_readbytespersector( fp, header );
	wii300_readsectorspercluster( fp, header );
	wii300_readmaxfilelength( fp, header );
	wii300_readtotalfiles( fp, header );
	wii300_readtotaldirs( fp, header );
	wii300_readtotalclusters( fp, header );
	wii300_readfreeclusters( fp, header );
	wii300_readserialnumber( fp, header );
	if( header->type == 8 )
		wii300_readheaderdescription( fp, header );
	wii300_readmedianame( fp, header );
	wii300_readfsname( fp, header );
}

struct wii300file *
wii300_readfile( FILE *fp )
{
	struct wii300file *filestruct;
	unsigned char charcount = 0;

	// Allocate some memory for the filestruct
	filestruct = (struct wii300file *) malloc( sizeof( struct wii300file ) );

	wii300_readfiletype( fp, filestruct );

#ifdef __DEBUG__
	printf( "filestruct->filetype: %u\n", filestruct->filetype );
#endif

	if( filestruct->filetype >= 66 && filestruct->filetype <= 71 )
		return filestruct;

	wii300_readfilename( fp, filestruct );

#ifdef __DEBUG__
	printf( "filestruct->filename: !%s!\n", filestruct->filename );
#endif

	wii300_readfilecreationdate( fp, filestruct );

	wii300_readfilesize( fp, filestruct );

#ifdef __DEBUG__
	printf( "filestruct->filesize: %u\n", filestruct->filesize );
#endif

	if( filestruct->filetype == 224 )
		wii300_readfiledescription( fp, filestruct );

	// Workaround for new filetype
	if( filestruct->filetype == 72 )
		fseek( fp, 5, SEEK_CUR );

#ifdef __DEBUG__
	if( filestruct->filetype == 224 )
		printf( "filestruct->filedescription: !%s!\n", filestruct->filedescription );
#endif

	return filestruct;
}

struct wii300dir *
wii300_readdir( FILE *fp )
{
	struct wii300dir *dirstruct;
	unsigned char charcount = 0;

	// Allocate some memory for the dirstruct
	dirstruct = (struct wii300dir *) malloc( sizeof( struct wii300dir ) );

	wii300_readdirtype( fp, dirstruct );

#ifdef __DEBUG__
	printf( "dirstruct->dirtype: %u\n", dirstruct->dirtype );
#endif

	wii300_readdirdesc( fp, dirstruct );

#ifdef __DEBUG__
	printf( "dirstruct->desc: %u\n", dirstruct->desc );
#endif

	wii300_readdirplaceintree( fp, dirstruct );

#ifdef __DEBUG__
	printf( "dirstruct->placeintree: %u\n", dirstruct->placeintree );
#endif

	wii300_readdirname( fp, dirstruct );

#ifdef __DEBUG__
	printf( "dirstruct->dirname: !%s!\n", dirstruct->dirname );
#endif

	wii300_readfilesindir( fp, dirstruct );

#ifdef __DEBUG__
	printf( "dirstruct->files: %u\n", dirstruct->files );
#endif

	wii300_readdircreationdate( fp, dirstruct );

#ifdef __DEBUG__
	printf( "dirstruct->timestamp: %u\n", dirstruct->timestamp );
	printf( "dirstruct->desc: %u\n", dirstruct->desc );
#endif

	if( dirstruct->desc == 1 )
		wii300_readdirdescription( fp, dirstruct );

#ifdef __DEBUG__
	if( dirstruct->desc == 1 )
		printf( "dirstruct->description: %s\n", dirstruct->description );
#endif

	return dirstruct;
}

void
wii300_seektocatalogname( FILE *fp, struct wii300header *header )
{
	// Skip until catalog name.
	fseek( fp, ( 2 * header->totaldisks ), SEEK_CUR );
}

void
wii300_seektodiskname( FILE *fp )
{
	// Skip until Diskname
	fseek( fp, 29L, SEEK_CUR );
}

void
wii300_seektomedianame( FILE *fp )
{
	// Skip until Medianame
	fseek( fp, 48L, SEEK_CUR );
}

void
wii300_skippastcatalogname( FILE *fp )
{
	unsigned int charcount = 0;

	wii300_seektocatalogname( fp, NULL );

	charcount = wii300_getplaintextlength( fp, 2 );

	wii300_skipplaintext( fp, charcount );
}

void
wii300_skippastdiskname( FILE *fp )
{
	unsigned int charcount = 0;

	wii300_seektodiskname( fp );

	charcount = wii300_getplaintextlength( fp, 2 );

	wii300_skipplaintext( fp, charcount );
}

void
wii300_skippastmedianame( FILE *fp )
{
	unsigned int charcount = 0;

	wii300_seektomedianame( fp );

	charcount = wii300_getplaintextlength( fp, 1 );

	wii300_skipplaintext( fp, charcount );
}

void
wii300_skippastfsname( FILE *fp )
{
	unsigned int charcount = 0;

	charcount = wii300_getplaintextlength( fp, 1 );

	wii300_skipplaintext( fp, charcount );
}

///////////////////////////////////////// FUNCTION FOR READING THE HEADER //////////////////////////////////////////
void
wii300_readtotaldisks( FILE *fp, struct wii300header *header )
{
	// Skip to totaldisks.
	fseek( fp, 16L, SEEK_SET );

	// We should really read 2 bytes here...
	header->totaldisks = wii300_getplaintextlength( fp, 2 );
}

void
wii300_readdisknumbers( FILE *fp, struct wii300header *header )
{
	int loop = 0;

	for( loop = 0; loop < header->totaldisks; loop++ )
		header->disknumbers[loop] = wii300_getplaintextlength( fp, 2 );
}

void
wii300_readcatalogname( FILE *fp, struct wii300header *header )
{
	unsigned int charcount = 0;

	charcount = wii300_getplaintextlength( fp, 2 );

	// Allocate place to store catalogname in
	header->catalogname = (char *) malloc( charcount + 1 );

	// Read the catalogname
	fread( header->catalogname, charcount, 1, fp );

	header->catalogname[ charcount ] = 0;

	// Seek forward to where the main header start
	fseek( fp, 8L, SEEK_CUR );
}

void
wii300_readdiskname( FILE *fp, struct wii300header *header )
{
	unsigned int charcount = 0;

	wii300_seektodiskname( fp );

 if( fgetc( fp ) != 0 )
 	fseek( fp, -1, SEEK_CUR );
	charcount = wii300_getplaintextlength( fp, 2 );
// fprintf(stderr, "diskname: %u\n", charcount );
	// Allocate place to store catalogname in
	header->diskname = (char *) malloc( charcount + 1 );

	// Read the catalogname
	fread( header->diskname, charcount, 1, fp );

	header->diskname[ charcount ] = 0;
}

void
wii300_readheaderformat( FILE *fp, struct wii300header *header )
{
	// Read the headerformat so if we should continue to read
	header->format = wii300_getplaintextlength( fp, 2 );
}

void
wii300_readdisknumber( FILE *fp, struct wii300header *header )
{
	// Read the current disknumber
	header->disknumber = wii300_getplaintextlength( fp, 2 );
}

void
wii300_readbytespersector( FILE *fp, struct wii300header *header )
{
	// Skip to bytespersector.
	fseek( fp, 1, SEEK_CUR );

	header->bytespersector = fgetc( fp );
}

void
wii300_readsectorspercluster( FILE *fp, struct wii300header *header )
{
	// Skip to sectorspercluster.
	fseek( fp, 2, SEEK_CUR );

	header->sectorspercluster = fgetc( fp );
}

void
wii300_readmaxfilelength( FILE *fp, struct wii300header *header )
{
	// Skip to maxfilelength
	fseek( fp, 3, SEEK_CUR );

	header->maxfilelength = fgetc( fp );
}

void
wii300_readtotalfiles( FILE *fp, struct wii300header *header )
{
	// Skip to totalfiles.
	fseek( fp, 3, SEEK_CUR );

	header->totalfiles = wii300_getplaintextlength( fp, 2 );
}

void
wii300_readtotaldirs( FILE *fp, struct wii300header *header )
{
	// Skip to totaldirs.
	fseek( fp, 6, SEEK_CUR );

	header->totaldirs = wii300_getplaintextlength( fp, 2 );
}

void
wii300_readtotalclusters( FILE *fp, struct wii300header *header )
{
	unsigned int field1 = 0;
	unsigned int field2 = 0;
	unsigned int field3 = 0;
	unsigned int field4 = 0;

	field2 = fgetc( fp );
	field3 = fgetc( fp );
	field4 = fgetc( fp );

	header->totalclusters = wii300_addsize( field1, field2, field3, field4 );
}

void
wii300_readfreeclusters( FILE *fp, struct wii300header *header )
{
	unsigned int field1 = 0;
	unsigned int field2 = 0;
	unsigned int field3 = 0;
	unsigned int field4 = 0;

	field1 = fgetc( fp );
	field2 = fgetc( fp );
	field3 = fgetc( fp );
	field4 = fgetc( fp );

	header->freeclusters = wii300_addsize( field1, field2, field3, field4 );
}

void
wii300_readserialnumber( FILE *fp, struct wii300header *header )
{
	// Skip to serialnumber.
	fseek( fp, 9, SEEK_CUR );

	header->serialnumber[0] = fgetc( fp );
	header->serialnumber[1] = fgetc( fp );
	header->serialnumber[2] = fgetc( fp );
	header->serialnumber[3] = fgetc( fp );

	// Seek two forward
	fseek( fp, 2, SEEK_CUR );

}

void
wii300_readheadertype( FILE *fp, struct wii300header *header )
{
	// Workaround to read extra character if there was a directory before
	unsigned char ch = fgetc( fp );

	if( ch != 3 && ch != 19 )
		fgetc( fp );

	// Skip to headertype.
	fseek( fp, 1, SEEK_CUR );

	header->type = wii300_getplaintextlength( fp, 1 );
}

void
wii300_readheaderdescription( FILE *fp, struct wii300header *header )
{
	unsigned int charcount = 0;

	charcount = wii300_getplaintextlength( fp, 2 );

#ifdef __DEBUG__
	printf( "headerdescription length: %u\n", charcount );
#endif

	// Allocate place to store headerdescription in
	header->description = (char *) malloc( charcount + 1 );

	// Read the headerdescription
	fread( header->description, charcount, 1, fp );

	header->description[ charcount ] = 0;

#ifdef __DEBUG__
	printf( "lastreadchar: %u\n", header->description[charcount-1] );
#endif
}

void
wii300_readmedianame( FILE *fp, struct wii300header *header )
{
	unsigned int charcount = 0;

	charcount = wii300_getplaintextlength( fp, 1 );

	// Read the catalogname
	fread( header->medianame, charcount, 1, fp );

	header->medianame[ charcount ] = 0;
}

void
wii300_readfsname( FILE *fp, struct wii300header *header )
{
	unsigned int charcount = 0;

	charcount = wii300_getplaintextlength( fp, 1 );

	// Read the catalogname
	fread( header->fsname, charcount, 1, fp );

	header->fsname[ charcount ] = 0;
}



void
wii300_freeheader( struct wii300header *header )
{
	free( header->catalogname );
	free( header->diskname );
	if( header->type == 226 )
		free( header->description );
	free( header );
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////// FUNCTION FOR READING THE CONTENT ///////////////////////////////////////////
void
wii300_readfiletype( FILE *fp, struct wii300file *filestruct )
{
	unsigned int type = 0;

	do {
		type = fgetc( fp );
	} while( type == 0 );

	filestruct->filetype = type;

	/* Different filetypes:
	 * 66 Directory
	 * 72,96 Plain file
	 * 224 File with description
	 */

	if( filestruct->filetype == 224 )
		filestruct->desc = 1;
	else
		filestruct->desc = 0;
}

void
wii300_readfilename( FILE *fp, struct wii300file *filestruct )
{
	unsigned int charcount = 0;

	// Skip to filename.
	fseek( fp, 1, SEEK_CUR );

	charcount = wii300_getplaintextlength( fp, 1 );

#ifdef __DEBUG__
	printf( "charcount: %u\n", charcount );
#endif

	// Read the filename.
	fread( filestruct->filename, charcount, 1, fp );

	filestruct->filename[ charcount ] = 0;
}

void
wii300_readfilecreationdate( FILE *fp, struct wii300file *filestruct )
{
	filestruct->timestamp = wii300_readdate( fp );
}

time_t
wii300_readdate( FILE *fp )
{
	time_t timestamp;
	struct tm *tblock;
	int field = 0;

	time( &timestamp );
	tblock = localtime( &timestamp );

	// Calculate date and put it into tblock.
	tblock->tm_sec = 0;
	tblock->tm_min = 0;
	tblock->tm_hour = 0;
	tblock->tm_mday = 0;
	tblock->tm_mon = 0;
	tblock->tm_year = 80;

	field = fgetc( fp );

	if( field >= 32 )
		tblock->tm_min += field / 32;
	tblock->tm_sec += ( field % 32 ) * 2;

	field = fgetc( fp );

	if( field >= 8 )
		tblock->tm_hour += field / 8;
	tblock->tm_min += ( field % 8 ) * 8;

	field = fgetc( fp );

	tblock->tm_mon += ( field / 32 ) - 1;
	tblock->tm_mday += field % 32;

	field = fgetc( fp );

	tblock->tm_year += field / 2;
	tblock->tm_mon += ( field % 2 ) * 8;

	return mktime( tblock );
}

void
wii300_readfilesize( FILE *fp, struct wii300file *filestruct )
{
	unsigned int field1 = 0;
	unsigned int field2 = 0;
	unsigned int field3 = 0;
	unsigned int field4 = 0;

	field1 = fgetc( fp );
	field2 = fgetc( fp );
	field3 = fgetc( fp );
	field4 = fgetc( fp );

	// Read the filesize.
	filestruct->filesize = wii300_addsize( field1, field2, field3, field4 );
}

void
wii300_readfiledescription( FILE * fp, struct wii300file *filestruct )
{
	unsigned int charcount = 0;

	// Skip to filedescription.
	fseek( fp, 1, SEEK_CUR );

	charcount = wii300_getplaintextlength( fp, 2 );

#ifdef __DEBUG__
	printf( "filedescription length: %u\n", charcount );
#endif

	// Seek to filedescription
	fseek( fp, 2, SEEK_CUR );

	// Allocate place to store filedescription in
	filestruct->filedescription = (char *) malloc( charcount + 1 );

	// Read the filedescription
	fread( filestruct->filedescription, charcount, 1, fp );

	filestruct->filedescription[ charcount ] = 0;

#ifdef __DEBUG__
	printf( "lastreadchar: %u\n", filestruct->filedescription[charcount-1] );
#endif
}

void
wii300_readdirtype( FILE *fp, struct wii300dir *dirstruct )
{
	do {
		dirstruct->dirtype = fgetc( fp );
	} while( dirstruct->dirtype == 0 );

	/* Different filetypes:
	 * 66-74 Directory
	 * 72,96 Plain file
	 * 224 File with description
	 */
}

void
wii300_readdirdesc( FILE *fp, struct wii300dir *dirstruct )
{
	unsigned int desc = 0;

	// Skip to filedescription.
	fseek( fp, 1, SEEK_CUR );

	desc = fgetc( fp );

	if( desc == 144 )
		dirstruct->desc = 1;
	else
		dirstruct->desc = 0;
}

void
wii300_readdirplaceintree( FILE *fp, struct wii300dir *dirstruct )
{
	// Skip to placeinttree.
	fseek( fp, 1, SEEK_CUR );

	dirstruct->placeintree = fgetc( fp );
}

void
wii300_readdirname( FILE *fp, struct wii300dir *dirstruct )
{
	unsigned int charcount = 0;

	charcount = wii300_getplaintextlength( fp, 2 );

	// Allocate place to store dirname in.
	dirstruct->dirname = (char *) malloc( charcount + 1 );

	// Read the dirname;
	fread( dirstruct->dirname, charcount, 1, fp );

	dirstruct->dirname[ charcount ] = 0;
}

void
wii300_readfilesindir( FILE *fp, struct wii300dir *dirstruct )
{
	// Skip to filesindir.
	fseek( fp, 12, SEEK_CUR );

	dirstruct->files = fgetc( fp );
}

void
wii300_readdircreationdate( FILE *fp, struct wii300dir *dirstruct )
{
	unsigned char ch;

	// Skip to dircreationdate
	fseek( fp, 15, SEEK_CUR );

	dirstruct->timestamp = wii300_readdate( fp );

	// Skip one char
	ch = fgetc( fp );
	if( ch < 39 || ch > 43 )
		fseek( fp, -1, SEEK_CUR );
}

void
wii300_readdirdescription( FILE * fp, struct wii300dir *dirstruct )
{
	unsigned int charcount = 0;

	// Skip to dirdescription.
	fseek( fp, 1, SEEK_CUR );

	charcount = wii300_getplaintextlength( fp, 2 );

	// Allocate place to store dirdescription in
	dirstruct->description = (char *) malloc( charcount + 1 );

	// Go to start of text.
	fseek( fp, 2, SEEK_CUR );

	// Read the dirdescription
	fread( dirstruct->description, charcount, 1, fp );

//	dirstruct->description[ charcount ] = 0;
	fseek( fp, 1, SEEK_CUR );

}

void
wii300_freefile( struct wii300file *filestruct )
{
	if( filestruct->filetype == 224 )
		free( filestruct->filedescription );
	free( filestruct );
}

void
wii300_freedir( struct wii300dir *dirstruct )
{
	free( dirstruct->dirname );
	if( dirstruct->desc == 1 )
		free( dirstruct->description );
	free( dirstruct );
}
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////// FUNCTIONS FOR BASIC OPERATIONS ///////////////////////////////////////////
unsigned int
wii300_getplaintextlength( FILE *fp, unsigned int fields )
{
	unsigned char ch = 0;
	unsigned char ch2 = 0;

	// When this function returns, FILE pointer stans on first character of the plaintext.
	if( fields == 2 )
	{
		ch = fgetc( fp );
		ch2 = fgetc( fp );
	}
	else if( fields = 1 )
		ch = fgetc( fp );
	else
		return 0;

	return wii300_addnormal( ch, ch2 );
}

void
wii300_skipplaintext( FILE *fp, unsigned char count )
{
	fseek( fp, count, SEEK_CUR );
}


unsigned int
wii300_addnormal( unsigned int first, unsigned int second )
{
	return first + ( second * 256 );
}

unsigned long
wii300_addsize( unsigned int first, unsigned int second, unsigned int third, unsigned fourth )
{
	return first + ( second * 256 ) + ( third * 65536 ) + ( fourth * 16777216 );
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
